using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

namespace CS_asynplex {
	enum VariableStatus {
		BASIC, 
		UNSELECTED, 
		SELECTED, 
		REJECTED
	}

	class VariableInfo {
		public VariableStatus status;
		public int basisWhereStatusChanged;

		public VariableInfo(VariableStatus status) {
			this.status = status;
			this.basisWhereStatusChanged = -1;
		}
	}

	class QueuedVariableRequest {
		public int unattractiveVariable;
		public int basis;
		public string processName;
		public string tagToSend;
	}

	class ColumnSelectionProcess : AsynplexProcess {
		//List<bool> selected = new List<bool>();
		List<VariableInfo> variables = new List<VariableInfo>();
		Queue<int> attractive = new Queue<int>();
		Queue<QueuedVariableRequest> queuedVariableRequests = new Queue<QueuedVariableRequest>();
		int m;
		int n;
		int kc;

		public ColumnSelectionProcess(int m, int n)
			: base("C") {
			this.m = m;
			this.n = n;
			this.kc = 0;
			this.variables.Add(null); // 1-based array; index 0 not used - TODO: This might not be the case in Phase I in Vanderbei?
			for (int i = 0; i < n; ++i) // nonbasic
				this.variables.Add(new VariableInfo(VariableStatus.UNSELECTED));
			for (int i = 0; i < m; ++i) // basic
				this.variables.Add(new VariableInfo(VariableStatus.BASIC));
		}

		private void installAttractiveCandidates(Message message) {
			AttractiveCandidatesMessage msg = (AttractiveCandidatesMessage)message.payload; // assumed to be sorted by attractiveness. index 0 is ki
			Debug.Assert(msg.attractiveVariables.Length >= 1);
			print("received " + msg.attractiveVariables.Length + " candidates from " + message.sender + ", which is at basis " + msg.basisNumber + "; currently at basis " + kc);
			if (msg.basisNumber > kc) {
				int installCount = 0;
				attractive = new Queue<int>();
				foreach (int cand in msg.attractiveVariables) {
					if (!((variables[cand].status == VariableStatus.SELECTED || variables[cand].status == VariableStatus.REJECTED) && variables[cand].basisWhereStatusChanged >= msg.basisNumber)) {
						++installCount;
						attractive.Enqueue(cand);
						variables[cand].status = VariableStatus.UNSELECTED;
						variables[cand].basisWhereStatusChanged = msg.basisNumber;
					}
					else {
						print("rejected x" + cand + " because it got " + variables[cand].status + " at basis " + variables[cand].basisWhereStatusChanged);
					}
				}
				print("Installed " + installCount + " candidates");
				kc = msg.basisNumber;
			}
		}

		private void waitForAndInstallAttractiveCandidates() {
			Message message;
			while ((message = receive("I*", "I9->C2")) == null) ;
			installAttractiveCandidates(message);
		}

		public override void run() {
			Message message;
			//waitForAndInstallAttractiveCandidates();
			while (!Program.Finished) {
				while (queuedVariableRequests.Count > 0 && attractive.Count > 0) {
					if (Program.Finished)
						return;
					int var = attractive.Dequeue();
					QueuedVariableRequest req = queuedVariableRequests.Dequeue();
					Debug.Assert(variables[var].status == VariableStatus.UNSELECTED);
					variables[var].status = VariableStatus.SELECTED;
					variables[var].basisWhereStatusChanged = kc;
					print("x" + var + " gets SELECTED (C5) at basis " + kc);
					send(req.processName, req.tagToSend, var);
				}
				while ((message = receive()) == null) {
					if (Program.Finished)
						return;
				}
				Debug.Assert(message.sender[0] == 'I');
				if (message.tag == "I8->C1") {
					int[] enteringLeavingBasis = (int[])message.payload;
					variables[enteringLeavingBasis[0]].status = VariableStatus.BASIC;
					variables[enteringLeavingBasis[0]].basisWhereStatusChanged = enteringLeavingBasis[2];
					variables[enteringLeavingBasis[1]].status = VariableStatus.UNSELECTED;
					variables[enteringLeavingBasis[1]].basisWhereStatusChanged = enteringLeavingBasis[2];
					print("x" + enteringLeavingBasis[1] + " left basis " + enteringLeavingBasis[2] + " (by " + message.sender + ") and got unselected");
				}
				else if (message.tag == "I9->C2") {
					installAttractiveCandidates(message);
					//queuedVariableRequests.Enqueue(new QueuedVariableRequest() { processName = message.sender, tagToSend = "C3->I11" });
					int var = attractive.Dequeue();
					Debug.Assert(variables[var].status == VariableStatus.UNSELECTED);
					variables[var].status = VariableStatus.SELECTED;
					variables[var].basisWhereStatusChanged = kc;
					print("x" + var + " gets SELECTED (C3) at basis " + kc);
					send(message.sender, "C3->I11", var);
				}
				else if (message.tag == "I4->C4") {
					int[] rejectedAndBasis = (int[])message.payload;
					if (rejectedAndBasis[0] != -1) {
						variables[rejectedAndBasis[0]].status = VariableStatus.REJECTED;
						variables[rejectedAndBasis[0]].basisWhereStatusChanged = rejectedAndBasis[1];
						print("x" + rejectedAndBasis[0] + " gets REJECTED (C5) at basis " + rejectedAndBasis[1]);
					}
					queuedVariableRequests.Enqueue(new QueuedVariableRequest() { processName = message.sender, tagToSend = "C5->I11" });
				}
				else {
					Debug.Assert(false, "C: Unexpected tag '" + message.tag + "'");
				}
			}
		}
	}
}
