using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Threading;

namespace CS_asynplex {
	struct AttractiveCandidate {
		public float reducedCost;
		public int index;
		public int column;
	}

	class AttractiveCandidatesMessage {
		public int basisNumber;
		public int[] attractiveVariables;
	}

	class BasisChangeMessage {
		public int enteringVar;
		public int leavingVar;
		public Matrix stepDirection;
		public float stepLength;
	}

	class IterationProcess : AsynplexProcess, IComparer<AttractiveCandidate> {
		int processIndex;
		int numIterationProcesses;
		int m, n;
		int ki;
		int currentEtaBasis = 0;
		const int reinversionPeriod = 5;
		Matrix N, B, originalBInverse;
		Matrix cNonbasic;
		Matrix cBasic;
		Matrix basicValues;
		Matrix origRhs;
		int[] nonbasicVars;
		int[] basicVars;
		bool[] basicStatus;
		int[] position;
		PrintControl printControl;
		const float EPS = 0.0000005f;
		List<Matrix> etaFile = new List<Matrix>();
		
		public void swap(ref int a, ref int b) {
			int temp = a;
			a = b;
			b = temp;
		}

		public IterationProcess(int processIndex, int numIterationProcesses, int m, int n, Matrix N, Matrix basicValues, Matrix cNonbasic)
			: base("I" + processIndex) {
			this.processIndex = processIndex;
			this.numIterationProcesses = numIterationProcesses;
			this.m = m;
			this.n = n;
			this.N = new Matrix(N);
			this.cNonbasic = new Matrix(cNonbasic);
			this.origRhs = new Matrix(basicValues);
			this.basicValues = new Matrix(basicValues);

			this.B = new Matrix(N.Rows, N.Rows, true, "B");
			this.originalBInverse = new Matrix(N.Rows, N.Rows, true, "B^I");
			this.cBasic = new Matrix(m, 1, "cBasic");
			this.nonbasicVars = new int[n];
			this.basicVars = new int[m];
			this.basicStatus = new bool[m + n + 1]; //1-indexed
			this.position = new int[m + n + 1]; //1-indexed

			for (int i = 1; i <= n; ++i) {
				this.nonbasicVars[i - 1] = i;
				this.basicStatus[i] = false;
				this.position[i] = i - 1;
			}
			for (int i = n + 1; i <= n + m; ++i) {
				this.basicVars[i - n - 1] = i;
				this.basicStatus[i] = true;
				this.position[i] = i - n - 1;
			}

			printControl = new PrintControl() { iterationCount = false, objectiveValue = true, pivotResults = true };
		}

		private static bool equalsEpsilon(float a, float b) {
			return Math.Abs(a - b) < EPS;
		}

		public int Compare(AttractiveCandidate a, AttractiveCandidate b) {
			if (equalsEpsilon(a.reducedCost, b.reducedCost))
				return a.index.CompareTo(b.index);
			return b.reducedCost.CompareTo(a.reducedCost);
		}

		private void BTRAN_calcPiTranspose(out Matrix piTranspose) {
			piTranspose = cBasic.transpose() * calcBInverseFromEtaFile(); // \pi^T
			piTranspose.Name = "pi^T";
			if (printControl.piTranspose) piTranspose.print();
		}

		private void PRICE_calcReducedCosts(Matrix piTranspose, out Matrix reducedCosts) {
			reducedCosts = cNonbasic.transpose() - piTranspose * N; // \hat{c}_N^T
			reducedCosts.Name = "reduced costs";
			if (printControl.reducedCosts) reducedCosts.print();
		}

		private void CHUZC_findAttractiveEntering(Matrix reducedCosts, out int enteringCol, out int enteringVar, out List<AttractiveCandidate> attractive) {
			attractive = new List<AttractiveCandidate>();
			for (int c = 0; c < reducedCosts.Cols; ++c) {
				if (reducedCosts[0, c] > 0) {
					attractive.Add(new AttractiveCandidate() { reducedCost = reducedCosts[0, c], index = nonbasicVars[c], column = c });
				}
			}
			attractive.Sort(this);

			if (printControl.pivotDecisions) {
				string output = "attractive candidates:";
				foreach (AttractiveCandidate cand in attractive) {
					output += " x" + cand.index + "(" + cand.reducedCost + ")";
				}
				print(output);
			}

			if (attractive.Count == 0) {
				print("Optimal");
				enteringCol = -1;
				enteringVar = -1;
				return;
			}
			else if (attractive.Count <= processIndex) {
				enteringCol = -1;
				enteringVar = -1;
				//throw new NotImplementedException();
				//TODO: What to do here?
			}
			else {
				enteringCol = attractive[processIndex].column;
				enteringVar = attractive[processIndex].index;
			}

			if (printControl.pivotResults) print("Entering variable: x" + enteringVar + " (column " + enteringCol + ")");
		}

		private void CHUZR_chooseLeaving(Matrix stepDirection, out int leavingRow, out int leavingVar, out float stepLength) {
			leavingRow = -1; // p
			float largestRatio = 0; // step length (\alpha)
			for (int r = 0; r < basicValues.Rows; ++r) {
				float ratio;
				if (stepDirection[r, 0] == 0 && basicValues[r, 0] == 0)
					ratio = 0;
				else
					ratio = stepDirection[r, 0] / basicValues[r, 0];
				if (ratio > 0 && (leavingRow == -1 || ratio > largestRatio || (ratio == largestRatio && basicVars[r] < basicVars[leavingRow]))) {
					if (printControl.pivotDecisions) print("Choosing " + (r+1) + " over " + (leavingRow+1) + " to leave; ratio is " + ratio);
					leavingRow = r;
					largestRatio = ratio;
				}
			}
			if (leavingRow == -1) {
				print("Unbounded");
				leavingVar = -1;
				stepLength = float.PositiveInfinity;
				return;
			}
			stepLength = 1 / largestRatio;
			leavingVar = basicVars[leavingRow];
			if (printControl.pivotResults) print("Leaving variable: x" + leavingVar + " (row " + leavingRow + ")");
			if (printControl.stepLength) print("Step length: " + stepLength);
		}

		private Matrix calcBInverseFromEtaFile() {
			Matrix result = originalBInverse;
			foreach (Matrix eta in etaFile) {
				result = eta * result;
			}
			if (etaFile.Count > 0 && printControl.inverseDiscrepancy) {
				float max = 0;
				Matrix discrepancy = result - B.invert();
				for (int r = 0; r < discrepancy.Rows; ++r) {
					for (int c = 0; c < discrepancy.Cols; ++c) {
						float absVal = Math.Abs(discrepancy[r, c]);
						if (absVal > max)
							max = absVal;
					}
				}
				Console.WriteLine("Largest discrepancy: {0}", max);
			}
			return result;
		}

		private void UPDATE_BASIS(int enteringCol, int leavingRow, Matrix stepDirection) {
			Matrix tempM = N.getColumn(enteringCol);
			N.setColumn(enteringCol, B.getColumn(leavingRow));
			B.setColumn(leavingRow, tempM);

			float dxi = stepDirection[leavingRow, 0];
			if (dxi == 0) throw new Exception();
			Matrix etaMatrix = new Matrix(B.Rows, B.Cols, true, "Eta" + etaFile.Count);
			for (int r = 0; r < etaMatrix.Rows; ++r) {
				if (r == leavingRow)
					etaMatrix[r, leavingRow] = 1 / dxi;
				else
					etaMatrix[r, leavingRow] = -stepDirection[r, 0] / dxi;
			}
			etaFile.Add(etaMatrix);
			if (printControl.etaFileSize) Console.WriteLine("{0} eta matrices", etaFile.Count);
			
			int enteringVar = nonbasicVars[enteringCol];
			int leavingVar = basicVars[leavingRow];
			
			float tempF = cNonbasic[enteringCol, 0];
			cNonbasic[enteringCol, 0] = cBasic[leavingRow, 0];
			cBasic[leavingRow, 0] = tempF;

			int tempI = nonbasicVars[enteringCol];
			nonbasicVars[enteringCol] = basicVars[leavingRow];
			basicVars[leavingRow] = tempI;

			basicStatus[enteringVar] = true;
			basicStatus[leavingVar] = false;
			tempI = position[enteringVar];
			position[enteringVar] = position[leavingVar];
			position[leavingVar] = tempI;

			if (printControl.variableLists) {
				Console.Write("Basic:    ");
				for (int i = 0; i < m; ++i)
					Console.Write(basicVars[i] + " ");
				Console.Write("\nNonbasic: ");
				for (int i = 0; i < n; ++i)
					Console.Write(nonbasicVars[i] + " ");
				Console.WriteLine();
			}
			if (printControl.iterationCount) print("\nCompleted iteration " + ki);
			if (printControl.iterationSeparator) print("\n\n");
		}

		private void UPRHS(int leavingRow, float stepLength, Matrix stepDirection) {
			basicValues -= stepLength * stepDirection;
			basicValues[leavingRow, 0] = stepLength;
			//basicValues = calcBInverseFromEtaFile() * origRhs;
			//basicValues.Name = "b";
			if (printControl.objectiveValue) {
				Matrix objectiveValue = cBasic.transpose() * basicValues;
				Debug.Assert(objectiveValue[0, 0] >= 0);
				print("Objective value: " + objectiveValue[0, 0]);
			}
		}

		private void applyIncomingBasisChanges() {
			Message message;
			while ((message = receive("I*", "I7->I2I3I10")) != null) {
				Debug.Assert(message.tag == "I7->I2I3I10");
				BasisChangeMessage msg = (BasisChangeMessage)message.payload;
				UPDATE_BASIS(position[msg.enteringVar], position[msg.leavingVar], msg.stepDirection);
				UPRHS(position[msg.leavingVar], msg.stepLength, msg.stepDirection);
				++ki;
			}
		}

		public override void run() {
			Message message;
			ki = 0;

			Matrix piTranspose;
			BTRAN_calcPiTranspose(out piTranspose);

			Matrix reducedCosts;
			PRICE_calcReducedCosts(piTranspose, out reducedCosts);

			int enteringCol, enteringVar;
			List<AttractiveCandidate> attractive;
			CHUZC_findAttractiveEntering(reducedCosts, out enteringCol, out enteringVar, out attractive);
			
			while (!Program.Finished) {
				while ((message = receive("V")) != null) {
					Debug.Assert(message.tag == "V2->I1");
					InverseMatrixMessage invMsg = (InverseMatrixMessage)message.payload;
					if (invMsg.basisNumber <= currentEtaBasis) { // Skip outdated matrices
						print("Got outdated B^I for basis " + invMsg.basisNumber + "; currently at basis " + currentEtaBasis + " / " + ki);
						continue;
					}
					if (invMsg.basisNumber > ki) {
						print("Got too new B^I for basis " + invMsg.basisNumber + "; currently at basis " + currentEtaBasis + " / " + ki);
						continue;
					}
					originalBInverse = invMsg.inverse;
					int removeCount = invMsg.basisNumber - currentEtaBasis;
					Debug.Assert(etaFile.Count >= removeCount);
					etaFile.RemoveRange(0, removeCount);
					currentEtaBasis += removeCount;
					print("Got new B^I for basis " + currentEtaBasis + "; eta file size is " + etaFile.Count);
				}
				applyIncomingBasisChanges();

				//FTRAN
				Matrix stepDirection = null;
				if (enteringCol != -1) {
					stepDirection = calcBInverseFromEtaFile() * N.getColumn(enteringCol); // \hat{a}_q
					stepDirection.Name = "step direction";
					if (printControl.stepDirection) stepDirection.print();
				}

				bool keepGoing;
				do {
					keepGoing = false;
					applyIncomingBasisChanges();

					//FTRAN
					if (enteringCol != -1) {
						stepDirection = calcBInverseFromEtaFile() * N.getColumn(enteringCol); // \hat{a}_q
						stepDirection.Name = "step direction";
						if (printControl.stepDirection) stepDirection.print();
					}

					Debug.Assert((enteringVar == -1 && enteringCol == -1) || enteringCol == position[enteringVar]);
					if (enteringVar == -1 || basicStatus[enteringVar] || reducedCosts[0, enteringCol] < 0) { //TODO: The pseudocode uses < - why perform a zero-advance pivot?
						send("C", "I4->C4", new int[]{enteringVar, ki});
					}
					else {
						send("R", "I5->R1", ki);
						while ((message = receive("R")) == null) {
							if (Program.Finished)
								return;
							Thread.Sleep(10);
						}
						Debug.Assert(message.tag == "R2->I6" || message.tag == "R3->I6");
						bool accepted = (bool)message.payload;
						if (accepted) {
							int leavingRow, leavingVar;
							float stepLength;
							CHUZR_chooseLeaving(stepDirection, out leavingRow, out leavingVar, out stepLength);
							if (leavingRow == -1) {
								Console.WriteLine("Unbounded");
								Program.Finished = true;
								Console.ReadLine();
								return;
							}
							for (int i = 0; i < numIterationProcesses; ++i) {
								if (i == processIndex) continue;
								BasisChangeMessage basisChangeMessage = new BasisChangeMessage() {enteringVar = enteringVar, leavingVar = leavingVar, stepDirection = stepDirection, stepLength = stepLength};
								send("I" + i, "I7->I2I3I10", basisChangeMessage);
							}
							send("V", "I8->V1", new int[] { enteringCol, leavingRow, });
							send("C", "I8->C1", new int[] { enteringVar, leavingVar, ki });
							UPDATE_BASIS(enteringCol, leavingRow, stepDirection);
							UPRHS(leavingRow, stepLength, stepDirection);
							++ki;
							BTRAN_calcPiTranspose(out piTranspose);
							PRICE_calcReducedCosts(piTranspose, out reducedCosts);
							CHUZC_findAttractiveEntering(reducedCosts, out enteringCol, out enteringVar, out attractive);
							if (attractive.Count == 0) {
								Console.WriteLine("Optimal");
								Program.Finished = true;
								Communication.printMessages();
								Console.ReadLine();
								return;
							}
							AttractiveCandidatesMessage msg = new AttractiveCandidatesMessage();
							msg.basisNumber = ki;
							msg.attractiveVariables = new int[attractive.Count];
							for (int i = 0; i < attractive.Count; ++i) {
								msg.attractiveVariables[i] = attractive[i].index;
								print(attractive[i].index + " is attractive");
							}
							send("C", "I9->C2", msg);
						}
						else {
							while ((message = receive("I*", "I7->I10")) == null) {
								if (Program.Finished)
									return;
								Thread.Sleep(10);
							}
							Debug.Assert(message.tag == "I7->I10", "Expected I7->I10, got " + message.tag);
							throw new NotImplementedException();
							//TODO: Apply basis change
							++ki;
							print("Keep going");
							keepGoing = true;
						}
					}
				} while (keepGoing);

				while ((message = receive("C")) == null) {
					if (Program.Finished)
						return;
					Thread.Sleep(10);
				}
				Debug.Assert(message.tag == "C3->I11" || message.tag == "C5->I11");
				applyIncomingBasisChanges();
				enteringVar = (int)message.payload;
				enteringCol = position[enteringVar];
				print("got suggested entering variable: x" + enteringVar);
			}
		}
	}
}
