using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CS_asynplex {
	class Matrix {
		private int rows;
		private int cols;
		private float[] data;
		private string name;

		public int Rows { get { return rows; } }
		public int Cols { get { return cols; } }
		public string Name { get { return name; } set { name = value; } }

		public Matrix(int rows, int cols) {
			if (rows <= 0 || cols <= 0)
				throw new ArgumentException();
			this.rows = rows;
			this.cols = cols;
			this.data = new float[rows * cols];
			for (int i = 0; i < rows * cols; ++i)
				this.data[i] = 0;
		}

		public Matrix(int rows, int cols, string name) : this(rows, cols) {
			this.name = name;
		}

		public Matrix(int rows, int cols, float[] data) {
			if (rows <= 0 || cols <= 0)
				throw new ArgumentException();
			this.rows = rows;
			this.cols = cols;
			this.data = new float[rows * cols];
			for (int i = 0; i < rows * cols; ++i)
				this.data[i] = data[i];
		}

		public Matrix(int rows, int cols, float[] data, string name) : this(rows, cols, data) {
			this.name = name;
		}

		public Matrix(int rows, int cols, bool identity) {
			if (rows <= 0 || cols <= 0)
				throw new ArgumentException();
			this.rows = rows;
			this.cols = cols;
			this.data = new float[rows * cols];
			for (int i = 0; i < rows * cols; ++i)
				this.data[i] = 0;
			if (identity && rows == cols) {
				for (int i = 0; i < rows; ++i) {
					this[i, i] = 1;
				}
			}
		}

		public Matrix(int rows, int cols, bool identity, string name) : this(rows, cols, identity) {
			this.name = name;
		}


		public Matrix(Matrix source) {
			this.rows = source.rows;
			this.cols = source.cols;
			this.data = new float[source.rows * source.cols];
			this.name = source.name;
			for (int i = 0; i < source.rows * source.cols; ++i)
				this.data[i] = source.data[i];
		}

		public Matrix copy() {
			return new Matrix(rows, cols, data, name);
		}

		public float this[int r, int c] {
			get {
				if (r < 0 || c < 0 || r >= rows || c >= cols)
					throw new ArgumentException("Index out of range");
				return data[r * cols + c];
			}
			set {
				if (r < 0 || c < 0 || r >= rows || c >= cols)
					throw new ArgumentException("Index out of range");
				data[r * cols + c] = value;
			}
		}

		public Matrix getColumn(int c) {
			if (c < 0 || c >= cols)
				throw new ArgumentException();
			Matrix column = new Matrix(rows, 1);
			for (int r = 0; r < rows; ++r)
				column[r, 0] = this[r, c];
			return column;
		}

		public void setColumn(int c, Matrix column) {
			if (c < 0 || c >= cols || column.cols != 1 || column.rows != rows)
				throw new ArgumentException();
			for (int r = 0; r < rows; ++r)
				this[r, c] = column[r, 0];
		}

		public void print() {
			Console.Write("=== {0} x {1}", rows, cols);
			if (name != null && name != "")
				Console.Write(" ({0})", name);
			Console.WriteLine(" ===");
			for (int r = 0; r < rows; ++r) {
				Console.Write(data[r * cols]);
				for (int c = 1; c < cols; ++c)
					Console.Write(" " + data[r * cols + c]);
				Console.WriteLine();
			}
			Console.WriteLine("======");
		}
		
		public static Matrix operator +(Matrix self, Matrix other) {
			if (self.rows != other.rows && self.cols != other.cols)
				throw new ArgumentException("Matrix sizes are not equal");
			Matrix result = new Matrix(self.rows, self.cols);
			for (int r = 0; r < self.rows; ++r) {
				for (int c = 0; c < self.cols; ++c) {
					result[r, c] = self[r, c] + other[r, c];
				}
			}
			return result;
		}

		public static Matrix operator -(Matrix self, Matrix other) {
			if (self.rows != other.rows && self.cols != other.cols)
				throw new ArgumentException("Matrix sizes are not equal");
			Matrix result = new Matrix(self.rows, self.cols);
			for (int r = 0; r < self.rows; ++r) {
				for (int c = 0; c < self.cols; ++c) {
					result[r, c] = self[r, c] - other[r, c];
				}
			}
			return result;
		}

		public static Matrix operator *(Matrix self, Matrix other) {
			if (self.cols != other.rows)
				throw new ArgumentException("Matrices are not compatible");
			Matrix result = new Matrix(self.rows, other.cols);
			for (int i = 0; i < self.rows; ++i) {
				for (int j = 0; j < other.cols; ++j) {
					float sum = 0;
					for (int k = 0; k < self.cols; ++k) {
						sum += self[i, k] * other[k, j];
					}
					if (Math.Abs(sum) < 0.00001f)
						sum = 0.0f;
					result[i, j] = sum;
				}
			}
			return result;
		}

		public static Matrix operator *(Matrix self, float factor) {
			Matrix result = new Matrix(self.rows, self.cols);
			for (int r = 0; r < self.rows; ++r) {
				for (int c = 0; c < self.cols; ++c) {
					result[r, c] = self[r, c] * factor;
				}
			}
			return result;
		}

		public static Matrix operator *(float factor, Matrix self) {
			return self * factor;
		}

		public Matrix transpose() {
			Matrix result = new Matrix(cols, rows);
			for (int r = 0; r < rows; ++r) {
				for (int c = 0; c < cols; ++c) {
					result[c, r] = this[r, c];
				}
			}
			return result;
		}

		public void multiplyRow(int row, float factor) {
			if (factor == 1) return;
			for (int j = 0; j < cols; ++j) {
				this[row, j] *= factor;
			}
		}

		public void addRows(int sourceRow, int destinationRow, float factor) {
			if (factor == 0) return;
			for (int j = 0; j < cols; ++j) {
				this[destinationRow, j] += this[sourceRow, j] * factor;
			}
		}

		public void swapRows(int firstRow, int secondRow) {
			if (firstRow == secondRow) return;
			for (int j = 0; j < cols; ++j) {
				float tmp = this[firstRow, j];
				this[firstRow, j] = this[secondRow, j];
				this[secondRow, j] = tmp;
			}
		}

		public Matrix invert() {
			if (rows != cols) throw new ArgumentException("Non-square matrices cannot be inverted");
			Matrix self = copy();
			Matrix inverse = new Matrix(rows, cols, true);
			for (int rc = 0; rc < cols; ++rc) {
				// Locate row with nonzero in this column
				int searchRow = rc;
				while (searchRow < rows && self[searchRow, rc] == 0)
					++searchRow;
				if (searchRow == rows)
					throw new ArgumentException("Matrix is singular");
				// Swap with current row; now the current row has nonzero in this column
				self.swapRows(rc, searchRow);
				inverse.swapRows(rc, searchRow);
				float factor = 1 / self[rc, rc];
				self.multiplyRow(rc, factor);
				inverse.multiplyRow(rc, factor);
				for (int r = 0; r < rows; ++r) {
					if (r == rc) continue;
					float innerFactor = -self[r, rc];
					self.addRows(rc, r, innerFactor);
					inverse.addRows(rc, r, innerFactor);
				}
			}
			return inverse;
		}
	}
}
