#ifndef ITERATIONPROCESS_H
#define ITERATIONPROCESS_H

#include "common/sparse.h"
#include "common/SafeVector.h"
#include "../PPU/payloadtypes.h"
#include "rpc.h"
#include "common/types.h"

/* The eta file seems to be stored as one sparse matrix, 
 * where each column represents one eta matrix. */
typedef struct {
	int iter;				/* Number of eta matrices */
	int size;				/* Number of entries in 'values' */
	SafeVector<TYPE> values;			/* Nonzero values */
	SafeVector<int> rowIndices;		/* Row indices - seem to be based on 'colPerm' */
	SafeVector<int> colPos;			/* Indices into 'values' and 'rowIndices' telling where each eta matrix column begins */
	SafeVector<int> newColLocations;	/* Column location for each eta matrix (leaving column for corresponding pivot) */
} EtaFile;

typedef struct {
	TYPE * yInternal;
	int * tagInternal;
	int currtagInternal;
} InternalData_SolveUsingLU;

typedef struct {
	TYPE * yInternal;
	int * tagInternal;
	int currtagInternal;
} InternalData_SolveTransposeUsingLU;

typedef struct {
	TYPE * a;
	int * tag;
	int * link;
	int currtag;
} InternalData_Gauss_Eta;

typedef struct {
	TYPE * a;
	int * tag;
	int currtag;
} InternalData_Gauss_Eta_T;

typedef struct {
	TYPE * a;
	int * tag;
	int * link;
	int currtag;
} InternalData_Nt_times_y;


typedef struct {
	int var;
	TYPE value;
} AttractiveCandidate;

class IterationProcess {
public:
	void setup(int index, int numIterationProcesses, SparseMatrix A, TYPE * b, TYPE * c, TYPE f);
	void run();

private:
	void expandEtaFile(int col_out);
	void installNewInverse();
	void FTRAN_calculateDualStepDirection(int col_out);
	void FTRAN_calculatePrimalStepDirection(int col_in);
	void CHUZR2_calculateStepLengths(int col_in, int col_out, TYPE * t_primalStepLength, TYPE * s_dualStepLength);
	void UPRHS_updatePrimalDualSolutions(int col_in, int col_out, TYPE primalStepLength, TYPE dualStepLength);
	void UPDATE_BASIS(int col_in, int col_out, int callRefactor);
	void applyIncomingBasisChange(BasisChangePayload * payload);
	void applyIncomingBasisChanges();
	void broadcastBasisChange(int fromBasis, int col_in, int col_out, int var_in, int var_out, TYPE primalStepLength, TYPE dualStepLength);
	void sendAttractiveCandidates(int basisNumber, const SafeVector<AttractiveCandidate> & attractive);
	int solveUsingLU(SafeSparseVector * y);
	int solveTransposeUsingLU(SafeSparseVector * y);
	int dbsolve(SafeVector<TYPE> & y);
	void Gauss_Eta(SafeSparseVector & dx_B);
	void Gauss_Eta_T(SafeSparseVector & vec);
	void Nt_times_y(SparseMatrix AT, const SafeVector<int> & basicflag, const SafeSparseVector & y, SafeSparseVector & yN);
	int CHUZR1_findLeaving_ratioTest(const SafeSparseVector & dy, const SafeVector<TYPE> & y);
	static TYPE recalculateObjectiveValues(const SafeVector<TYPE> & c, const SafeVector<TYPE> & x_B, const SafeVector<int> & basics);
	static void CHUZC_findAttractiveCandidates(const SafeVector<int> & columnToVariableMapping, const SafeVector<TYPE> & x, SafeVector<AttractiveCandidate> & attractive);
	bool isBasic(int var);
	bool isNonbasic(int var);
	bool columnForVariable(int var);
	void makeIdentityMatrix(FactoredInverse * inv, int size);
	
	int index;
	int numIterationProcesses;
	Communication comm;
	
	SparseMatrix A;
	SparseMatrix AT;
	SafeVector<TYPE> b;	/* right-hand side */
	SafeVector<TYPE> c;	/* objective coefficients */
	TYPE f;		/* objective function shift */
	SafeVector<TYPE> x_B;	/* primal basics */
	SafeVector<TYPE> y_N;	/* dual nonbasics */
	
	int currentBasis;
	
	FactoredInverse factoredInverse;
	
	int     e_iter; /* number of iterations since last refactorization */
	
	EtaFile etaFile;
	
	SafeSparseVector dyN_dualStepDir;
	SafeSparseVector dxB_primalStepDir;
	SafeSparseVector tempVector;
	
	SafeVector<int> basics;		/* list of basic variable indices */
	SafeVector<int> nonbasics;	/* list of non-basic variable indices */
	SafeVector<int> basicflag;	/* negative if nonbasic, nonnegative if basic */
	
	InternalData_SolveUsingLU internalData_SolveUsingLU;
	InternalData_SolveTransposeUsingLU internalData_SolveTransposeUsingLU;
	InternalData_Gauss_Eta internalData_Gauss_Eta;
	InternalData_Gauss_Eta_T internalData_Gauss_Eta_T;
	InternalData_Nt_times_y internalData_Nt_times_y;
	
	double cumtime; //TODO: This one was used both by the functions that were moved to invertprocessor.c and to interationprocess.c
	double ocumtime;
};

#endif
