/*********************************************************************/
/***    Copyright (c) Robert J. Vanderbei, 1994                    ***/
/***    All Rights Reserved                                        ***/
/*********************************************************************/

/* #include <stdio.h> */
#include "macros.h"
/* #include <sigfpe.h> */

#define	HEADER 0
#define	NAME   1
#define	ROWS   2
#define	COLS   3
#define	RHS    4
#define	RNGS   5
#define	BNDS   6
#define	QUADS  7
#define	END    8

#define	UNSET  0
#define	PRIMAL 1
#define	DUAL   2

#define FINITE   0x1
#define INFINITE 0x2
#define UNCONST  0x4
#define FREEVAR   0x1
#define BDD_BELOW 0x2
#define BDD_ABOVE 0x4
#define BOUNDED   0x8

#define	LP_OPEN_MAX 20 /* max #	lp problems open at once */

typedef	struct lp_prob {
	int m;		/* number of rows */
	int n;		/* number of columns */
	int nz;		/* number of nonzeros */
	TYPE *A;	/* pointer to array of nonzero values in A */
	int *iA;	/* pointer to array of corresponding row indices */
	int *kA;	/* pointer to array of indices into A (and iA)
				indicating where each new column of A begins */
	TYPE *b;	/* pointer to array containing right-hand side */
	TYPE *c;	/* pointer to array containing objective function */
	TYPE f;	/* fixed adjustment to objective function */
	TYPE *r;	/* pointer to array containing range vector */
	TYPE *l;	/* pointer to array containing lower bounds */
	TYPE *u;	/* pointer to array containing upper bounds */
	int *varsgn;	/* array indicating which variables were declared to
				be non-positive	*/
	char **rowlab;	/* array of strings containing row labels */
	char **collab;	/* array of strings containing column labels */

	int qnz;	/* number of nonzeros in lower triangle	of Q */
	TYPE *Q;	/* pointer to array of nonzero values of Q */
	int *iQ;	/* pointer to array of corresponding row indices */
	int *kQ;	/* pointer to array of indices into Q (and iQ)
				indicating where each new column of Q begins */

	TYPE *At;	/* pointer to array of nonzero values in At */
	int *iAt;	/* pointer to array of corresponding row indices */
	int *kAt;	/* pointer to array of indices into At (and iAt) */

	int *bndmark;	/* pointer to array of bound marks */
	int *rngmark;	/* pointer to array of range marks */

	TYPE *w;	/* pointer to array containing primal surpluses	*/
	TYPE *x;	/* pointer to array containing primal solution */
	TYPE *y;	/* pointer to array containing dual solution */
	TYPE *z;	/* pointer to array containing dual slacks */
	TYPE *p;	/* pointer to array containing range slacks */
	TYPE *q;	/* pointer to array containing dual range slacks */
	TYPE *s;	/* pointer to array containing dual for	ub slacks */
	TYPE *t;	/* pointer to array containing upper bound slacks */
	TYPE *v;	/* pointer to array containing dual for	range (w) */
	TYPE *g;	/* pointer to array containing lower bound slacks */
	TYPE *ub;	/* pointer to array containing shifted upper bounds */

	int max;	/* max = -1, min = 1 */
	TYPE inftol;	/* infeasibility tolerance */
	TYPE inftol2;	/* infeasibility for stopping rule */
	int sf_req;	/* significant figures requested */
	int itnlim;	/* iteration limit */
	double timlim;	/* time limit */
	int verbose;	/* level of verbosity */
	TYPE epssol;	/* epsilon tolerance in f/b solve */
	TYPE epsnum;	/* epsilon tolerance in num fact */
	TYPE epscdn;	/* epsilon tolerance for conditioning */
	TYPE stablty;	/* mixing factor for stability */
	int method;	/* reordering method */
	int dense;	/* threshold for dense columns/rows */
	int pdf;	/* order to favor primal (ADA^T) or dual (A^TDA) */
	char name[15];	/* string containing problem name */
	char obj[11];	/* string containing objective function	name */
	char rhs[11];	/* string containing right-hand	side name */
	char ranges[11];/* string containing range set name */
	char bounds[11];/* string containing bound set name */

	int *tier;	/* tier for factorization priorities */

	char **param;	/* array of strings containing user parameters */
	int np;		/* number of user parameters */
 
	int  (*stopping_rule)(void *);/* pointer to stopping	rule fcn */
	void (*init_vars)(void *);    /* pointer to initialization fcn */

	void (*h_init)(void *);       /* pointer to initialization hook fcn */
	void (*h_update)(void *);     /* pointer to update hook fcn */
	void (*h_step)(void *);       /* pointer to step hook fcn */

	int    iter;	    /* current iteration number	*/
	double elaptime;    /* elapsed time */
	TYPE pres;	    /* primal residual (i.e. infeasibility) */
	TYPE dres;	    /* dual   residual (i.e. infeasibility) */
	int    sigfig;	    /* significant figures */
	TYPE primal_obj;  /* primal objective	value */
	TYPE dual_obj;    /* dual   objective	value */
} LP;

LP	*openlp(void);

void	readlp(
	int	argc,
	char	*argv[],
	LP	*lp
);

int	solvelp(
	LP	*lp
);

void	closelp(
	LP	*lp
);

int	getparam(
	char *label0
);

void	my_exit(
	int num,
	char *str
);

extern void message();

TYPE recalculateObjectiveValues(
	TYPE	*a,
	int	*ja,
	TYPE  *x,
	int	na
);

TYPE dotprod(		/* inner product between n-vectors x and y */
	TYPE	*x,
	TYPE	*y,
	int	n
);

TYPE dotprodm(	/* inner product between n-vectors x and y */
	TYPE	*x,
	TYPE	*y,
	int	n,
	int	*mask,
	int	val
);

void smx(		/* y = sparse matrix (A,kA,iA) times x */
	int	m,
	int	n,
	TYPE	*A,
	int	*kA,
	int	*iA,
	TYPE	*x,
	TYPE	*y
);

void smtx(		/* y = x times sparse matrix (A,kA,iA) */
	int	m,
	int	n,
	TYPE	*A,
	int	*kA,
	int	*iA,
	TYPE	*x,
	TYPE	*y
);

void transposeMatrix(		/* (kAt,iAt,At)	= transpose of (kA,iA,A) */
	int	m,
	int	n,
	int	*kA,
	int	*iA,
	TYPE	*A,
	int	*kAt,
	int	*iAt,
	TYPE	*At
);

TYPE maxv(		/* compute componentwise maximum of n-vector x */
	TYPE *x,
	int	n
);

void	writelp(
	LP	*lp,
	char	*fname
);

void	writesol(
	LP	*lp,
	char	fname[]
);

void inv_num(
	LP 	*lp,	/* pointer to the linear program's data */
        TYPE  *dn,    /* diagonal matrix for upper-left  corner */
        TYPE  *dm     /* diagonal matrix for lower-right corner */
);

int solve(
	LP 	*lp,
	TYPE	*Dn,
	TYPE	*Dm,
        TYPE  *c,
        TYPE  *b
);

int deflt_st_rule(
    void *
);

void deflt_init(
    void *
);

void deflt_hook(
    void *
);

void inv_clo(void);

char *my_strdup(
	char	*s1
);

double cputimer(
);

void nl_init( 
    void *vlp 
);

void nl_update( 
    void *vlp 
);

void nlobjterm( 
        void (*func)( TYPE *z, TYPE *pval, TYPE *grad, TYPE **hessian),                           /* function that computes val, grad, hessian at z */
        int k,             /* number of arguments for func() */
        char **rowlabs,    /* row labels for arugments to func() */
        TYPE *argshifts  /* shifts to the arguments to func() */
);

void nlconstr( 
        void (*func)( TYPE *z, TYPE *pval, TYPE *grad, TYPE **hessian),                           /* function that computes val, grad, hessian at z */
	TYPE rhs,
        int k,             /* number of arguments for func() */
        char **rowlabs,    /* row labels for arugments to func() */
        TYPE *argshifts  /* shifts to the arguments to func() */
);

void nlobjterm0( 
        void (*func)( TYPE *z, TYPE *pval, TYPE *grad, TYPE **hessian),                           /* function that computes val, grad, hessian at z */
        int k,             /* number of arguments for func() */
        char **rowlabs,    /* row labels for arugments to func() */
        TYPE *argshifts  /* shifts to the arguments to func() */
);

void nlconstr0( 
        void (*func)( TYPE *z, TYPE *pval, TYPE *grad, TYPE **hessian),                           /* function that computes val, grad, hessian at z */
	TYPE rhs,
        int k,             /* number of arguments for func() */
        char **rowlabs,    /* row labels for arugments to func() */
        TYPE *argshifts  /* shifts to the arguments to func() */
);
