static int useAsynplex = 1;
/*****************************************************************************

 Implementation of the
 2 phase (Dual then Primal) Simplex Method
 R. Vanderbei, 3 October 1994

 Solves problems in the form:

 T
 max c x

 A x  = b
 x >= 0

 A is an m by N matrix (it is convenient to reserve n for
 the difference N-m).  Artificial variables have been
 added (hence N > m).  One may assume that the last
 m columns of A are invertible and hence can be used as
 a starting basis.

 ******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <pthread.h>
#include "print.h"
#include "timer.h"
#include "communication.h"
#include "invertprocessor.h"
#include "iterationprocess.h"
#include "basischangemanager.h"
#include "columnselectionmanager.h"

#ifdef QuadPrec
#include "Quad.h"
#define TYPE Quad
#else
#define high(x) (x)
#endif

#include "../common/linalg.h"
#include "lu.h"
#include "../common/myalloc.h"
#include "../common/macros.h"

#define EPS_MATRIX 1.0e-8
#define EPS_MU_COEFFICIENTS 1.0e-12
#define EPS_MU_OPTIMAL 1.0e-12 //3.0e-4
#define MAX_ITER 1000000

#define EPS  1.0e-8
#define EPS1 1.0e-8
#define EPS2 1.0e-12
#define EPS3 1.0e-12

#define numIterationProcesses 2

//TODO: Can we find a drand48() implementation for windows?
#ifdef WIN32
#define drand48() (((double)rand())/((double)RAND_MAX))
#endif

TYPE recalculateObjectiveValues(TYPE *c, TYPE *x_B, int *basics, int m);

void Nt_times_y(SparseMatrix AT, int *basicflag, SparseVector y,
		SparseVector * yN);

int ratio_test(TYPE *dy, int *idy, int ndy, TYPE *y);

int pick_neg(int m, TYPE *x);

int solver(SparseMatrix A, TYPE *b, /* right-hand side */
	TYPE *c, /* objective coefficients */
	TYPE f, /* objective function shift */
	TYPE *x, /* primal solution (output) */
	TYPE *y, /* dual solution (output) */
	TYPE *w, /* primal slacks (output) */
	TYPE *z /* dual slacks (output) */
) {
	TYPE *x_B; /* primal basics */
	TYPE *y_N; /* dual nonbasics */

	SparseVector dyN_dualStepDir;
	SparseVector dxB_primalStepDir;
	SparseVector vector;

	int *basics; /* list of basic variable indices */
	int *nonbasics; /* list of non-basic variable indices */
	int *basicflag; /* neg if nonbasic, pos if basic */

	int col_in; /* entering column; index in 'nonbasics' */
	int col_out; /* leaving column; index in 'basics' */

	int iter = 0; /* number of iterations */
	int i, j, k, m, n, N, v = 0;

	TYPE s, t;
	float primal_obj;

	int from_scratch;
	
	SparseMatrix AT;
	
	pthread_t invertProcessor;
	pthread_t basisChangeManager;
	pthread_t columnSelectionManager;
	IterationProcess iterationProcesses[numIterationProcesses];
	
	unsigned long long startTime = rdtsc();

	/*******************************************************************
	 * For convenience, we put...
	 *******************************************************************/
	m = A.rows;
	N = A.cols;
	n = N - m;

	/*******************************************************************
	 * Add slack variables.  We assume the calling routine allocated
	 * enough memory.
	 *******************************************************************/

	i = 0;
	k = A.colPos[n];
	for (j = n; j < N; j++) {
		c[j] = 0.0;
		A.values[k] = 1.0;
		A.rowIndices[k] = i;
		i++;
		k++;
		A.colPos[j + 1] = k;
	}
	A.numNonzeroes = k;
	
	/*******************************************************************
	 * Read in the Data and initialize the common memory sites.
	 *******************************************************************/

	MALLOC( x_B, m, TYPE );
	MALLOC( dxB_primalStepDir.values, m, TYPE );
	MALLOC( y_N, n, TYPE );
	MALLOC( dyN_dualStepDir.values, n, TYPE );
	MALLOC( vector.values, m, TYPE );
	MALLOC( nonbasics, n, int );
	MALLOC( basics, m, int );
	MALLOC( basicflag, N, int );
	MALLOC( dxB_primalStepDir.rowIndices, m, int );
	MALLOC( dyN_dualStepDir.rowIndices, n, int );
	MALLOC( vector.rowIndices, m, int );
	MALLOC( AT.values, A.numNonzeroes, TYPE );
	MALLOC( AT.rowIndices, A.numNonzeroes, int );
	MALLOC( AT.colPos, m+1, int );
	vector.rows = m;
	dxB_primalStepDir.rows = m;
	dyN_dualStepDir.rows = n;
	
	
	
	/****************************************************************
	 *  Threading initialization.              				        *
	 ****************************************************************/
	if (useAsynplex) {
		initPrintFromProcess();
		initCommunication();
		setupInvertProcessor(A, numIterationProcesses);
		setupColumnSelectionManager(m, n);
		pthread_create(&invertProcessor, NULL, runInvertProcessor, NULL);
		pthread_create(&basisChangeManager, NULL, runBasisChangeManager, NULL);
		pthread_create(&columnSelectionManager, NULL, runColumnSelectionManager, NULL);
		printf("Using %d iteration processes\n", numIterationProcesses);
		for (i = 0; i < numIterationProcesses; ++i) {
			setupIterationProcess(&iterationProcesses[i], i, numIterationProcesses, A, b, c, f);
			pthread_create(&iterationProcesses[i].thread, NULL, runIterationProcess, &iterationProcesses[i]);
		}
		goto waitForAsynplex;
	}
	
	
	
	/****************************************************************
	 *  Initialization.              				    *
	 ****************************************************************/

	transposeSparseMatrix(A, &AT);

	for (j = 0; j < n; j++) {
		nonbasics[j] = j;
		basicflag[j] = -j - 1;
		y_N[j] = MAX(c[j],1.0); /* to force dual feasibility */
		y_N[j] += drand48(); /* to ensure nondegeneracy   */
	}

	for (i = 0; i < m; i++) {
		basics[i] = n + i;
		basicflag[n + i] = i;
		x_B[i] = b[i];
	}

	lufac(m, A.colPos, A.rowIndices, A.values, basics, 0);

	dbsolve(m, x_B); /* could be done explicitly in terms of bounds/ranges */

	/****************************************************************
	 * 	Begin Phase I (i.e., dual simplex method)
	 ****************************************************************/
		
	printf("m = %d,n = %d,nz = %d\n", m, N, A.numNonzeroes);
	printf(
			"---------------------------------------------------------------------------\n"
				"          |   Primal      |        |                           arithmetic  \n"
				"  Iter    |  Obj Value    |   mu   |   nonz(L)     nonz(U)     operations  \n"
				"- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
	fflush(stdout);

	/****************************************************************
	 * 	Main loop                                                   *
	 ****************************************************************/

	for (iter = 0; iter < MAX_ITER; iter++) {

		/*************************************************************
		 * STEP 6: Print statistics and factor/refactor               *
		 *************************************************************/

		primal_obj = recalculateObjectiveValues(c, x_B, basics, m) + f;
		if (iter % 100 == 0) {
		printf("%8d   %14.7e     NA    \n", iter, high(primal_obj));
		fflush(stdout);
		}

		/*************************************************************
		 * STEP 1: Pick most negative basic primal                    *
		 *************************************************************/

		col_out = pick_neg(m, x_B);
		if (col_out == -1)
			break; /* ready for Phase II */

		/*************************************************************
		 *                          -1  T                             *
		 * STEP 2: Compute dy  = -(B  N) e                            *
		 *                   N            i			     *
		 *         where i = col_out                                  *
		 *************************************************************/

		vector.values[0] = -1.0;
		vector.rowIndices[0] = col_out;
		vector.numNonzeroes = 1;

		solveTransposeUsingLU(&vector);

		Nt_times_y(AT, basicflag, vector, &dyN_dualStepDir);

		/*************************************************************
		 * STEP 3: Ratio test to find entering column                 *
		 *************************************************************/
		col_in = ratio_test(dyN_dualStepDir.values, dyN_dualStepDir.rowIndices,
				dyN_dualStepDir.numNonzeroes, y_N);

		if (col_in == -1) {
			return 2; /* INFEASIBLE */
		}

		/*************************************************************
		 *                        -1                                  *
		 * STEP 4: Compute dx  = B  N e                               *
		 *                   B         j                              *
		 *                                                            *
		 *************************************************************/

		j = nonbasics[col_in];
		for (i = 0, k = A.colPos[j]; k < A.colPos[j + 1]; i++, k++) {
			dxB_primalStepDir.values[i] = A.values[k];
			dxB_primalStepDir.rowIndices[i] = A.rowIndices[k];
		}
		dxB_primalStepDir.numNonzeroes = i;

		solveUsingLU(&dxB_primalStepDir);

		/*************************************************************
		 *                                                            *
		 * STEP 5: Put       t = x /dx                                *
		 *                        i   i                               *
		 *                   s = y /dy                                *
		 *                        j   j                               *
		 *************************************************************/

		/* this is inefficient - it should be fixed */
		for (k = 0; k < dxB_primalStepDir.numNonzeroes; k++)
			if (dxB_primalStepDir.rowIndices[k] == col_out)
				break;

		t = x_B[col_out] / dxB_primalStepDir.values[k];

		/* this is inefficient - it should be fixed */
		for (k = 0; k < dyN_dualStepDir.numNonzeroes; k++)
			if (dyN_dualStepDir.rowIndices[k] == col_in)
				break;

		s = y_N[col_in] / dyN_dualStepDir.values[k];

		/*************************************************************
		 * STEP 7: Set y  = y  - s dy                                 *
		 *              N    N       N                                *
		 *                                                            *
		 *             y  = s                                         *
		 *              i                                             *
		 *                                                            *
		 *             x  = x  - t dx                                 *
		 *              B    B       B                                *
		 *                                                            *
		 *             x  = t                                         *
		 *              j                                             *
		 *************************************************************/

		for (k = 0; k < dyN_dualStepDir.numNonzeroes; k++) {
			j = dyN_dualStepDir.rowIndices[k];
			y_N[j] -= s * dyN_dualStepDir.values[k];
		}

		y_N[col_in] = s;

		for (k = 0; k < dxB_primalStepDir.numNonzeroes; k++) {
			i = dxB_primalStepDir.rowIndices[k];
			x_B[i] -= t * dxB_primalStepDir.values[k];
		}

		x_B[col_out] = t;
		/*************************************************************
		 * STEP 8: Update basis                                       *
		 *************************************************************/

		i = basics[col_out];
		j = nonbasics[col_in];
		basics[col_out] = j;
		nonbasics[col_in] = i;
		basicflag[i] = -col_in - 1;
		basicflag[j] = col_out;

		from_scratch = refactor(m, A.colPos, A.rowIndices, A.values, basics,
				col_out, v);
	} /* End of Phase I */

	printf("%8d   %14.7e     NA    \n", iter, high(primal_obj));
	printf("End of Phase I \n");

	/****************************************************************
	 *   Restore objective function by setting                       *
	 *                -1  T                                          *
	 *         y  = (B  N) c  - c                                    *
	 *          N           B    N                                   *
	 ****************************************************************/

	vector.numNonzeroes = 0;
	for (i = 0; i < m; i++) {
		if (ABS(c[basics[i]]) > EPS) {
			vector.values[vector.numNonzeroes] = c[basics[i]];
			vector.rowIndices[vector.numNonzeroes] = i;
			vector.numNonzeroes++;
		}
	}

	solveTransposeUsingLU(&vector);

	Nt_times_y(AT, basicflag, vector, &dyN_dualStepDir);

	for (j = 0; j < n; j++)
		y_N[j] = -c[nonbasics[j]];
	for (k = 0; k < dyN_dualStepDir.numNonzeroes; k++) {
		j = dyN_dualStepDir.rowIndices[k];
		y_N[j] += dyN_dualStepDir.values[k];
	}

	/****************************************************************
	 *   Begin Phase II (I.e., primal simplex method)                *
	 ****************************************************************/

	for (; iter < MAX_ITER; iter++) {

		/*************************************************************
		 * STEP 6: Print statistics and factor/refactor               *
		 *************************************************************/

		primal_obj = recalculateObjectiveValues(c, x_B, basics, m) + f;
		if (iter % 100 == 0) {
		printf("%8d   %14.7e     NA    \n", iter, high(primal_obj));
		fflush(stdout);
		}

		/*************************************************************
		 * STEP 1: Pick most negative nonbasic dual                   *
		 *************************************************************/

		col_in = pick_neg(n, y_N);
		if (col_in == -1)
			break; /* optimal */

		/*************************************************************
		 *                        -1                                  *
		 * STEP 2: Compute dx  = B  N e                               *
		 *                   B         j                              *
		 *         where j = col_in                                   *
		 *************************************************************/

		j = nonbasics[col_in];
		for (i = 0, k = A.colPos[j]; k < A.colPos[j + 1]; i++, k++) {
			dxB_primalStepDir.values[i] = A.values[k];
			dxB_primalStepDir.rowIndices[i] = A.rowIndices[k];
		}
		dxB_primalStepDir.numNonzeroes = i;
				
		solveUsingLU(&dxB_primalStepDir);
		

		/*************************************************************
		 * STEP 3: Ratio test to find leaving column                  *
		 *************************************************************/

		col_out = ratio_test(dxB_primalStepDir.values,
				dxB_primalStepDir.rowIndices, dxB_primalStepDir.numNonzeroes,
				x_B);

		if (col_out == -1)
			return 1; /* UNBOUNDED */
		
		
		/*************************************************************
		 *                          -1  T                             *
		 * STEP 4: Compute dy  = -(B  N) e                            *
		 *                   N            i			     *
		 *         where i = col_out                                  *
		 *************************************************************/

		vector.values[0] = -1.0;
		vector.rowIndices[0] = col_out;
		vector.numNonzeroes = 1;

		solveTransposeUsingLU(&vector);

		Nt_times_y(AT, basicflag, vector, &dyN_dualStepDir);

		/*************************************************************
		 * STEP 5: Put       t = x /dx                                *
		 *                        i   i                               *
		 *                   s = y /dy                                *
		 *                        j   j                               *
		 *************************************************************/

		/* this is inefficient - it should be fixed */
		for (k = 0; k < dxB_primalStepDir.numNonzeroes; k++)
			if (dxB_primalStepDir.rowIndices[k] == col_out)
				break;

		t = x_B[col_out] / dxB_primalStepDir.values[k];

		/* this is inefficient - it should be fixed */
		for (k = 0; k < dyN_dualStepDir.numNonzeroes; k++)
			if (dyN_dualStepDir.rowIndices[k] == col_in)
				break;

		s = y_N[col_in] / dyN_dualStepDir.values[k];
		/*************************************************************
		 * STEP 7: Set y  = y  - s dy                                 *
		 *              N    N       N                                *
		 *                                                            *
		 *             y  = s                                         *
		 *              i                                             *
		 *                                                            *
		 *             x  = x  - t dx                                 *
		 *              B    B       B                                *
		 *                                                            *
		 *             x  = t                                         *
		 *              j                                             *
		 *************************************************************/

		for (k = 0; k < dyN_dualStepDir.numNonzeroes; k++) {
			j = dyN_dualStepDir.rowIndices[k];
			y_N[j] -= s * dyN_dualStepDir.values[k];
		}

		y_N[col_in] = s;

		for (k = 0; k < dxB_primalStepDir.numNonzeroes; k++) {
			i = dxB_primalStepDir.rowIndices[k];
			x_B[i] -= t * dxB_primalStepDir.values[k];
		}

		x_B[col_out] = t;

				
		/*************************************************************
		 * STEP 8: Update basis                                       *
		 *************************************************************/

		i = basics[col_out];
		j = nonbasics[col_in];
		basics[col_out] = j;
		nonbasics[col_in] = i;
		basicflag[i] = -col_in - 1;
		basicflag[j] = col_out;

		from_scratch = refactor(m, A.colPos, A.rowIndices, A.values, basics,
				col_out, v);
	} /* End of Phase II */

	primal_obj = recalculateObjectiveValues(c, x_B, basics, m) + f;
	printf("%8d   %14.7e     NA    \n", iter, high(primal_obj));
	printf("End of Phase II \n");

	/****************************************************************
	 * 	Transcribe solution to x vector and dual solution to y      *
	 ****************************************************************/

	for (j = 0; j < N; j++)
		x[j] = 0.0;
	for (i = 0; i < m; i++)
		x[basics[i]] = x_B[i];
	//printTypeArray(x, N);

	for (j = 0; j < N; j++)
		y[j] = 0.0;
	for (i = 0; i < n; i++)
		y[nonbasics[i]] = y_N[i];

	/****************************************************************
	 * 	Split out slack variables and shift dual variables.
	 ****************************************************************/

	for (j = 0; j < n; j++)
		z[j] = y[j];
	for (i = 0; i < m; i++) {
		y[i] = y[n + i];
		w[i] = x[n + i];
	}

	/****************************************************************
	 * 	Free work space                                             *
	 ****************************************************************/

	FREE( vector.values );
	FREE( vector.rowIndices );
	FREE( x_B );
	FREE( y_N );
	FREE( dxB_primalStepDir.values );
	FREE(dxB_primalStepDir.rowIndices );
	FREE( dyN_dualStepDir.values );
	FREE(dyN_dualStepDir.rowIndices );
	FREE( nonbasics );
	FREE( basics );
	
	if (useAsynplex) {
waitForAsynplex:
		for (i = numIterationProcesses - 1; i >= 0; --i) {
			pthread_join(iterationProcesses[i].thread, NULL);
		}
		pthread_join(invertProcessor, NULL);
		pthread_join(columnSelectionManager, NULL);
		pthread_join(basisChangeManager, NULL);
	}
	printf("Time spent (minus input handling): %llu\n", rdtsc() - startTime);
	
	return 0;

} /* End of solver */

void Display_Solution(int m, int *basics, TYPE *X) {
	int i;

	printf("SOLUTION:\n\n");
	for (i = 0; i < m; i++)
		printf("  X[%d] = %lf\n", basics[i], high(X[i]));
}

void Nt_times_y(SparseMatrix AT, int *basicflag, SparseVector y,
		SparseVector * yN) {
	int i, j, jj, k, kk;

	static TYPE *a = NULL;
	static int *tag = NULL;
	static int *link = NULL;
	static int currtag = 1;

	if (a == NULL)
		MALLOC( a, AT.rows, TYPE);
	if (tag == NULL)
		CALLOC( tag, AT.rows, int);
	if (link == NULL) {
		CALLOC(link, AT.rows+2, int);
		link++;
	}

	jj = -1;
	for (k = 0; k < y.numNonzeroes; k++) {
		i = y.rowIndices[k];
		for (kk = AT.colPos[i]; kk < AT.colPos[i + 1]; kk++) {
			j = AT.rowIndices[kk];
			if (basicflag[j] < 0) {
				if (tag[j] != currtag) {
					a[j] = 0.0;
					tag[j] = currtag;
					link[jj] = j;
					jj = j;
				}
				a[j] += y.values[k] * AT.values[kk];
			}
		}
	}
	link[jj] = AT.rows;
	currtag++;

	k = 0;
	for (jj = link[-1]; jj < AT.rows; jj = link[jj]) {
		if (ABS(a[jj]) > EPS_MATRIX) {
			yN->values[k] = a[jj];
			yN->rowIndices[k] = -basicflag[jj] - 1;
			k++;
		}
	}
	yN->numNonzeroes = k;
}

int ratio_test(TYPE *dy, int *idy, int ndy, TYPE *y) {
	int j, jj = -1, k;
	TYPE min = HUGE_VAL;

	for (k = 0; k < ndy; k++) {
		if (dy[k] > EPS1) {
			j = idy[k];
			if (y[j] / dy[k] < min) {
				min = y[j] / dy[k];
				jj = j;
			}
		}
	}

	return jj;
}

TYPE recalculateObjectiveValues(TYPE *c, TYPE *x_B, int *basics, int m) {
	int i;
	TYPE prod = 0.0;

	for (i = 0; i < m; i++) {
		prod += c[basics[i]] * x_B[i];
	}

	return prod;
}

int pick_neg(int m, TYPE *x) {
	int i;
	int col = -1;
	TYPE min = -EPS2;

	for (i = 0; i < m; i++) {
		if (x[i] < min) {
			min = x[i];
			col = i;
		}
	}
	return col;
}
