#include "payloadtypes.h"
#include "iterationprocess.h"
#include "invertprocessor.h"
#include "sparse.h"
#include "communication.h"
#include "print.h"
#include "util.h"
#include "sparse.h"
#include "../common/macros.h"
#include "../common/myalloc.h"
#include "../common/tree.h"
#include "../common/linalg.h"
#include "../common/message.h"
#include <stdio.h>
#include <math.h>
#include <assert.h>
using namespace std;

#define EPS_MATRIX 1.0e-8
#define EPSSOL	1.0e-5   /* Zero tolerance for consistent eqns w/dep rows */
#define E_N		200
#define E_NZ	20000
#define SOLVE_USING_LU_EPS	1.0e-14 /* EPS from lueta.c */
#define ITERATION_EPS		1.0e-8	/* EPS from 2phase.c */
#define PICK_NEG_EPS		1.0e-10	/* EPS2 from 2phase.c */
#define GAUSS_ETA_EPS		1.0e-14	/* EPS1 from lueta.c */
#define RATIO_TEST_ETA		1.0e-10	/* EPS1 from 2phase.c */
#define MAX_ITER	1000000

//TODO: Can we find a drand48() implementation for windows?
#ifdef WIN32
#define drand48() (((double)rand())/((double)RAND_MAX))
#endif

#ifdef QuadPrec
#include "Quad.h"
#define TYPE Quad
#else
#define high(x) (x)
#endif


// Ugly to have this here, but putting it in genericvectors caused heavy circular dependency problems. Should me moved.
#define VECTOR_NAME AttractiveCandidateVector
#define VECTOR_TYPE AttractiveCandidate
#define VECTOR_INIT initACV
#define VECTOR_RESIZE resizeACV
#define VECTOR_FREE freeACV
#define VECTOR_APPEND appendACV
#define VECTOR_GET getACV
#define VECTOR_SET setACV
#define VECTOR_REMOVE_AT removeAtACV
#define VECTOR_COPY copyACV
#include "../common/genericvector_headers.h"
#include "../common/genericvector_source.h"
#undef VECTOR_COPY
#undef VECTOR_REMOVE_AT
#undef VECTOR_SET
#undef VECTOR_GET
#undef VECTOR_APPEND
#undef VECTOR_FREE
#undef VECTOR_RESIZE
#undef VECTOR_INIT
#undef VECTOR_TYPE
#undef VECTOR_SHORT_NAME
#undef VECTOR_NAME




static int iterationProcess_dbsolve(IterationProcess * proc, SafeVector<TYPE> & y);
static int iterationProcess_solveUsingLU(IterationProcess * proc, SafeSparseVector * y);
static int iterationProcess_solveTransposeUsingLU(IterationProcess * proc, SafeSparseVector * y);
static void iterationProcess_Gauss_Eta(IterationProcess * proc, SafeSparseVector & dx_B);
static void iterationProcess_Gauss_Eta_T(IterationProcess * proc, SafeSparseVector & vec);
//static void iterationProcess_lu_clo(IterationProcess * proc);
static void iterationProcess_Nt_times_y(IterationProcess * proc, SparseMatrix AT, const SafeVector<int> & basicflag, const SafeSparseVector & y, SafeSparseVector & yN);
static int CHUZR1_findLeaving_ratioTest(IterationProcess * proc, const SafeSparseVector & dy, const SafeVector<TYPE> & y);
static TYPE iterationProcess_recalculateObjectiveValues(const SafeVector<TYPE> & c, const SafeVector<TYPE> & x_B, const SafeVector<int> & basics);
static void CHUZC_findAttractiveCandidates(const SafeVector<int> & columnToVariableMapping, const SafeVector<TYPE> & x, AttractiveCandidateVector * attractive);

static void send(IterationProcess * proc, const char * receiver, const char * tag, void * payload) {
	Message msg = {proc->name, receiver, tag, payload};
	sendMessage(msg);
}

#define receiveFromAnyone() receiveMessageFromAnyone(proc.name)

#define receive(sender) receiveMessage(proc.name, sender)

#define receiveWithTag(sender, tag) receiveMessageWithTag(proc.name, sender, tag)


int enablePrint = 1;

//http://ou800doc.sco.com/cgi-bin/info2html?(gcc.info)Macro%2520Varargs&lang=en
#define print(formatString, ...) if (enablePrint) { printFromProcess(proc.name, formatString , ## __VA_ARGS__); }
#define printp(formatString, ...) if (enablePrint) { printFromProcess(proc->name, formatString , ## __VA_ARGS__); }


#define streq(a, b) (!strcmp(a, b))

#define isBasic(var) (proc.basicflag[(var)] >= 0)
#define isNonbasic(var) (proc.basicflag[(var)] < 0)

#define columnForVariable(var) (proc.basicflag[(var)] >= 0 ? proc.basicflag[(var)] : -proc.basicflag[(var)] - 1)
#define columnForVariableP(var) (proc->basicflag[(var)] >= 0 ? proc->basicflag[(var)] : -proc->basicflag[(var)] - 1)


void makeIdentityMatrix(FactoredInverse * inv, int size) {
	int i;
	makeZeroSparseMatrix(&inv->L, size, size);
	makeZeroSparseMatrix(&inv->LT, size, size);
	makeZeroSparseMatrix(&inv->U, size, size);
	makeZeroSparseMatrix(&inv->UT, size, size);
	// All permutation vectors are filled with 0, size-1, ..., 1
	inv->colPerm.resize(size);
	inv->colPerm.at(0) = 0;
	for (i = 1; i < size; ++i) {
		inv->colPerm.at(i) = size - i;
	}
	inv->invColPerm = inv->colPerm;
	inv->rowPerm = inv->colPerm;
	inv->invRowPerm = inv->colPerm;
	// Diagonal is only 1s
	inv->diagU.resize(size);
	for (i = 0; i < size; ++i) {
		inv->diagU[i] = 1;
	}
	inv->rank = size;
	inv->basisNumber = 0;
}

void setupIterationProcess(IterationProcess * process, int index, int numIterationProcesses, SparseMatrix A, TYPE * b, TYPE * c, TYPE f) {
	int i;
	int m = A.rows;
	int N = A.cols;
//	int n = N - m;
	char * name = (char*)malloc(16); // Long enough for any conceivable index number
	sprintf(name, "I%d", index);
	process->name = name;
	process->index = index;
	process->numIterationProcesses = numIterationProcesses;
	copySparseMatrix(A, &process->A);
	for (i = 0; i < m; ++i)
		process->b.push_back(b[i]);
	for (i = 0; i < N; ++i)
		process->c.push_back(c[i]);
	process->f = f;
	
	process->currentBasis = 0;
	
	makeIdentityMatrix(&process->factoredInverse, m);
	
	// Reset eta file
	//TODO: Make eta file expandable
	//TODO: Do partial reset only
	process->etaFile.iter = 0; // Updated in refactor
	process->etaFile.size = 0; // Updated at the end of solveUsingLU
	process->etaFile.colPos.push_back(0);
	
	process->tempVector.setRows(m);
	
	process->internalData_SolveUsingLU.currtagInternal = 1;
	CALLOC( process->internalData_SolveUsingLU.yInternal, m, TYPE);
	CALLOC( process->internalData_SolveUsingLU.tagInternal, m, int);
	
	process->internalData_SolveTransposeUsingLU.currtagInternal = 1;
	CALLOC( process->internalData_SolveTransposeUsingLU.yInternal, m, TYPE);
	CALLOC( process->internalData_SolveTransposeUsingLU.tagInternal, m, int);
	
	process->internalData_Gauss_Eta.currtag = 1;
	CALLOC( process->internalData_Gauss_Eta.a, m, TYPE);
	CALLOC( process->internalData_Gauss_Eta.tag, m, int);
	CALLOC( process->internalData_Gauss_Eta.link, m + 2, int);
	process->internalData_Gauss_Eta.link++; //TODO: Call free() on link-1 !
	
	process->internalData_Gauss_Eta_T.currtag = 1;
	CALLOC( process->internalData_Gauss_Eta_T.a, m, TYPE);
	CALLOC( process->internalData_Gauss_Eta_T.tag, m, int);
	
	process->internalData_Nt_times_y.currtag = 1;
	MALLOC( process->internalData_Nt_times_y.a, N, TYPE);
	CALLOC( process->internalData_Nt_times_y.tag, N, int);
	CALLOC( process->internalData_Nt_times_y.link, N + 2, int);
	process->internalData_Nt_times_y.link++; //TODO: Call free() on link-1 !
	
	process->cumtime = 0;
	process->ocumtime = 0;
	
	process->iterationProcessNames = (char**)malloc(numIterationProcesses * sizeof(char *));
	for (i = 0; i < numIterationProcesses; ++i) {
		process->iterationProcessNames[i] = (char*)malloc(16);
		sprintf(process->iterationProcessNames[i], "I%d", i);
	}
}

static void expandEtaFile(IterationProcess * proc, int col_out) {
	proc->etaFile.newColLocations.push_back(col_out);
	proc->etaFile.iter++;
}

//TODO: Make 'peek' functionality so that we can ignore all but the newest inverse
// Repeatedly checks for messages from V with the tag "V2->I1".
// Such a message will contain a more up-to-date LU factorisation of the basis matrix than what this iteration process already has.
// As long as a message is found, the new inverse will be installed (proc->factoredInverse will be replaced by the payload of the message)
// and a suitable amount of entries from the front of the eta file will be removed.
void installNewInverse(IterationProcess * proc) {
	Message message;
	FactoredInverse * inversePayload;
	int i;
	
	while ((message = receiveMessage(proc->name, "V")).sender) {
		assert(!strcmp(message.tag, "V2->I1"));
		inversePayload = (FactoredInverse*)message.payload;
		//TODO: Restructure loops
		if (inversePayload->basisNumber <= proc->factoredInverse.basisNumber) {
			printp("Got outdated  B^I for basis %d; currently at basis %d", inversePayload->basisNumber, proc->factoredInverse.basisNumber);
		}
		else if (inversePayload->basisNumber > proc->currentBasis) {
			printp("Got too new B^I for basis %d; process is at basis %d; inverse is at basis %d", inversePayload->basisNumber, proc->currentBasis, proc->factoredInverse.basisNumber);
		}
		else {
			int numEtaEntriesToRemove;
			int newPos;
			printp("Got acceptable B^I for basis %d; process is at basis %d; inverse is at basis %d", inversePayload->basisNumber, proc->currentBasis, proc->factoredInverse.basisNumber);
			printp("installNewInverse is modifying eta file");
			numEtaEntriesToRemove = inversePayload->basisNumber - proc->factoredInverse.basisNumber;
			newPos = proc->etaFile.colPos.at(numEtaEntriesToRemove);
			printp("%d %d", numEtaEntriesToRemove, newPos);
			proc->etaFile.size -= newPos;
			proc->etaFile.iter -= numEtaEntriesToRemove;
			// TODO: Can we use some sort of queue structure to make this more efficient?
			for (i = 0; i < proc->etaFile.size; ++i) {
				proc->etaFile.rowIndices.at(i) = proc->etaFile.rowIndices.at(i + newPos);
				proc->etaFile.values.at(i) = proc->etaFile.values.at(i + newPos);
			}
			proc->etaFile.rowIndices.resize(proc->etaFile.size);
			proc->etaFile.values.resize(proc->etaFile.size);
			for (i = 0; i < proc->etaFile.iter + 1; ++i) {
				proc->etaFile.colPos.at(i) = proc->etaFile.colPos.at(i + numEtaEntriesToRemove) - newPos;
			}
			proc->etaFile.colPos.resize(proc->etaFile.iter + 1);
			assert(proc->etaFile.colPos.at(0) == 0);
			for (i = 0; i < proc->etaFile.iter; ++i) {
				proc->etaFile.newColLocations.at(i) = proc->etaFile.newColLocations.at(i + numEtaEntriesToRemove);
			}
			proc->etaFile.newColLocations.resize(proc->etaFile.iter);
			// Free the space occupied by the previous matrix, copy the message payload and free the original payload
			freeSparseMatrix(&proc->factoredInverse.L);
			freeSparseMatrix(&proc->factoredInverse.LT);
			freeSparseMatrix(&proc->factoredInverse.U);
			freeSparseMatrix(&proc->factoredInverse.UT);
			proc->factoredInverse = *inversePayload;
			printp("basis: %d", proc->factoredInverse.basisNumber);
		}
		delete inversePayload;
	}
}

/*************************************************************
 *                            -1  T                           *
 * STEP 2/4: Compute dy  = -(B  N) e                          *
 *                     N            i			              *
 *           where i = col_out                                *
 *************************************************************/
// proc->dyN_dualStepDir will be updated.
// The objective function coefficients will change by dyN_dualStepDir * s_dualStepLength (which will be calculated later).
void FTRAN_calculateDualStepDirection(IterationProcess * proc, int col_out) {
	proc->tempVector.resize(1);
	proc->tempVector.value(0) = -1.0;
	proc->tempVector.rowIndex(0) = col_out;
	iterationProcess_solveTransposeUsingLU(proc, &proc->tempVector);
	iterationProcess_Nt_times_y(proc, proc->AT, proc->basicflag, proc->tempVector, proc->dyN_dualStepDir);
}

/*************************************************************
 *                          -1                                *
 * STEP 4/2: Compute dx  = B  N e                             *
 *                     B         j                            *
 *           where j = col_in                                 *
 *************************************************************/
// proc->dxB_primalStepDir will be updated.
// The right hand side will change by dxB_primalStepDir * t_primalStepLength (which will be calculated later).
// This function calls iterationProcess_solveUsingLU(), which as a side effect will partially update the eta file - but etaFile.iter won't be touched, so the change won't be visible until expandEtaFile() is called.
void FTRAN_calculatePrimalStepDirection(IterationProcess * proc, int col_in) {
	int i, j, k;
	j = proc->nonbasics[col_in];
	proc->dxB_primalStepDir.clear();
	for (i = 0, k = proc->A.colPos[j]; k < proc->A.colPos[j + 1]; i++, k++) {
		proc->dxB_primalStepDir.append(proc->A.values[k], proc->A.rowIndices[k]);
	}
	iterationProcess_solveUsingLU(proc, &proc->dxB_primalStepDir);
}

/*************************************************************
 *                                                            *
 * STEP 5: Put       t = x /dx                                *
 *                        i   i                               *
 *                   s = y /dy                                *
 *                        j   j                               *
 *************************************************************/
// Given the leaving and entering columns, computes the factors that dxB_primalStepDir and dyN_dualStepDir must be multiplied by.
void CHUZR2_calculateStepLengths(IterationProcess * proc, int col_in, int col_out, TYPE * t_primalStepLength, TYPE * s_dualStepLength) {
	int k;
	
	/* this is inefficient - it should be fixed */
	for (k = 0; k < proc->dxB_primalStepDir.numNonzeroes(); k++)
		if (proc->dxB_primalStepDir.rowIndex(k) == col_out)
			break;
	
	*t_primalStepLength = proc->x_B[col_out] / proc->dxB_primalStepDir.value(k);
	
	/* this is inefficient - it should be fixed */
	for (k = 0; k < proc->dyN_dualStepDir.numNonzeroes(); k++)
		if (proc->dyN_dualStepDir.rowIndex(k) == col_in)
			break;
	
	*s_dualStepLength = proc->y_N[col_in] / proc->dyN_dualStepDir.value(k);
}

/*************************************************************
 * STEP 7: Set y  = y  - s dy                                 *
 *              N    N       N                                *
 *                                                            *
 *             y  = s                                         *
 *              i                                             *
 *                                                            *
 *             x  = x  - t dx                                 *
 *              B    B       B                                *
 *                                                            *
 *             x  = t                                         *
 *              j                                             *
 *************************************************************/
void UPRHS_updatePrimalDualSolutions(IterationProcess * proc, int col_in, int col_out, TYPE primalStepLength, TYPE dualStepLength) {
	int i, j, k;
	for (k = 0; k < proc->dyN_dualStepDir.numNonzeroes(); k++) {
		j = proc->dyN_dualStepDir.rowIndex(k);
		proc->y_N[j] -= dualStepLength * proc->dyN_dualStepDir.value(k);
	}
	
	proc->y_N[col_in] = dualStepLength;
	
	for (k = 0; k < proc->dxB_primalStepDir.numNonzeroes(); k++) {
		i = proc->dxB_primalStepDir.rowIndex(k);
		proc->x_B[i] -= primalStepLength * proc->dxB_primalStepDir.value(k);
	}
	
	proc->x_B[col_out] = primalStepLength;
}

/*************************************************************
 * STEP 8: Update basis                                       *
 *************************************************************/
void UPDATE_BASIS(IterationProcess * proc, int col_in, int col_out, int callRefactor) {
	int i, j;
	i = proc->basics[col_out];
	j = proc->nonbasics[col_in];
	proc->basics[col_out] = j;
	proc->nonbasics[col_in] = i;
	proc->basicflag[i] = -col_in - 1;
	proc->basicflag[j] = col_out;
	//if (callRefactor)
		//iterationProcess_refactor(proc, proc->basics, col_out);
}

void applyIncomingBasisChange(IterationProcess * proc, Message message) {
	BasisChangePayload * payload = (BasisChangePayload*)message.payload;
	int i;
	printp("Applying incoming basis change: to %d", payload->fromBasis + 1);
	assert(payload->fromBasis == proc->currentBasis);
	proc->dxB_primalStepDir = payload->primalStepDirection;
	proc->dyN_dualStepDir = payload->dualStepDirection;
	for (i = 0; i < payload->etaFileSizeDiff; ++i) {
		proc->etaFile.values.push_back(payload->newEtaFileValues.at(i));
		proc->etaFile.rowIndices.push_back(payload->newEtaFileRowIndices.at(i));
	}
	proc->etaFile.size += payload->etaFileSizeDiff;
	proc->etaFile.colPos.push_back(proc->etaFile.colPos.at(proc->etaFile.iter) + payload->etaFileSizeDiff);
	proc->etaFile.newColLocations.push_back(payload->leavingRow);
	proc->etaFile.iter++;
	
	UPDATE_BASIS(proc, payload->enteringCol, payload->leavingRow, 0);
	UPRHS_updatePrimalDualSolutions(proc, payload->enteringCol, payload->leavingRow, payload->primalStepLength, payload->dualStepLength);
	++proc->currentBasis;
	delete payload;
}

void applyIncomingBasisChanges(IterationProcess * proc) {
	Message message;
	while ((message = receiveMessageWithTag(proc->name, "I*", "I7->I2I3I10")).sender) {
		assert(!strcmp(message.tag, "I7->I2I3I10"));
		applyIncomingBasisChange(proc, message);
	}
}

// The eta file must be fully updated, including etaFile.iter, before this function is called.
void broadcastBasisChange(IterationProcess * proc, int fromBasis, int col_in, int col_out, int var_in, int var_out, TYPE primalStepLength, TYPE dualStepLength) {
	InvProcPivotPayload * invProcPayload = (InvProcPivotPayload*)malloc(sizeof(InvProcPivotPayload)); // Will be free'd by InvProc
	ColSelPivotPayload * colSelPayload = (ColSelPivotPayload*)malloc(sizeof(ColSelPivotPayload)); // Will be free'd by ColSelProc
	int i, j, k;
	BasisChangePayload * basisChangePayload;
	Message * messages = new Message[proc->numIterationProcesses - 1];
	Message * message = messages;
	for (i = 0; i < proc->numIterationProcesses; ++i) {
		if (i == proc->index) continue;
		basisChangePayload = new BasisChangePayload;
		basisChangePayload->enteringCol = col_in;
		basisChangePayload->leavingRow = col_out;
		basisChangePayload->fromBasis = fromBasis;
		basisChangePayload->primalStepLength = primalStepLength;
		basisChangePayload->dualStepLength = dualStepLength;
		basisChangePayload->primalStepDirection = proc->dxB_primalStepDir;
		basisChangePayload->dualStepDirection = proc->dyN_dualStepDir;
		basisChangePayload->etaFileSizeDiff = proc->etaFile.colPos.at(proc->etaFile.iter) - proc->etaFile.colPos.at(proc->etaFile.iter - 1);
		for (j = 0, k = proc->etaFile.colPos.at(proc->etaFile.iter - 1); j < basisChangePayload->etaFileSizeDiff; ++j, ++k) {
			basisChangePayload->newEtaFileValues.push_back(proc->etaFile.values.at(k));
			basisChangePayload->newEtaFileRowIndices.push_back(proc->etaFile.rowIndices.at(k));
		}
		printp("Sending basis change message to I%d (fromBasis == %d)", i, fromBasis);
		message->sender = proc->name;
		message->receiver = proc->iterationProcessNames[i];
		message->tag = "I7->I2I3I10";
		message->payload = basisChangePayload;
		++message;
	}
	sendAll(messages, proc->numIterationProcesses - 1);
	delete [] messages;
	invProcPayload->enteringCol = col_in;
	invProcPayload->leavingRow = col_out;
	send(proc, "V", "I8->V1", invProcPayload);
	colSelPayload->enteringVar = var_in;
	colSelPayload->leavingVar = var_out;
	colSelPayload->basisNumber = proc->currentBasis;
	send(proc, "C", "I8->C1", colSelPayload);
}

void sendAttractiveCandidates(IterationProcess * proc, int basisNumber, AttractiveCandidateVector * attractive) {
	int i;
	AttractiveCandidatesMessage * payload = new AttractiveCandidatesMessage;
	payload->basisNumber = basisNumber;
	for (i = 0; i < attractive->size; ++i) {
		payload->attractiveVariables.push_back(getACV(*attractive, i).var);
		//printp("%d is attractive", getACV(*attractive, i).col);
	}
	send(proc, "C", "I9->C2", payload);
}

void * runIterationProcess(void * parameters) {
	IterationProcess proc = *(IterationProcess*)parameters;
		
	SafeSparseVector tempVector;
	
	int col_in; /* entering column; index in 'nonbasics' */
	int col_out; /* leaving column; index in 'basics' */
	int var_out;
	int var_in;

	int iter = 0; /* number of iterations */
	int i, j, k, m, n, N;

	TYPE s_dualStepLength, t_primalStepLength;
	float primal_obj;
	
	TYPE *x; /* primal solution (output) */
	TYPE *y; /* dual solution (output) */
	TYPE *w; /* primal slacks (output) */
	TYPE *z; /* dual slacks (output) */
	
	Message message;
	InvProcPivotPayload * pivotPayload;
	ChuzrOfferPayload * chuzrOfferPayload;
	ChuzrAcceptancePayload * chuzrAcceptancePayload;
	CandidateIsUnattractivePayload * unattractivePayload;

	AttractiveCandidateVector attractive;
	
	int keepGoing;
	int finished;
	int accepted;
	
	seedRandomGeneratorForThisThread();
	
	/*******************************************************************
	 * For convenience, we put...
	 *******************************************************************/
	m = proc.A.rows;
	N = proc.A.cols;
	n = N - m;
	print("m = %d, n = %d, N = %d, nz = %d", m, n, N, proc.A.numNonzeroes);
	
	/*******************************************************************
	 * Read in the Data and initialize the common memory sites.
	 *******************************************************************/
	
	proc.x_B.resize(m);
	proc.dxB_primalStepDir.setRows(m);
	proc.y_N.resize(n);
	proc.dyN_dualStepDir.setRows(n);
	tempVector.setRows(m);
	proc.nonbasics.resize(n);
	proc.basics.resize(m);
	proc.basicflag.resize(N);
	MALLOC( proc.AT.values, proc.A.numNonzeroes, TYPE );
	MALLOC( proc.AT.rowIndices, proc.A.numNonzeroes, int );
	MALLOC( proc.AT.colPos, m+1, int );
	nullifySparseMatrix(&proc.AT);
	initACV(&attractive);
	
	/****************************************************************
	 *  Initialization.              				    *
	 ****************************************************************/
	
	transposeSparseMatrix(proc.A, &proc.AT);
	
	for (j = 0; j < n; j++) {
		proc.nonbasics[j] = j;
		proc.basicflag[j] = -j - 1;
		proc.y_N[j] = MAX(proc.c[j],1.0); /* to force dual feasibility */
		proc.y_N[j] += drand48(); /* to ensure nondegeneracy   */
	}

	for (i = 0; i < m; i++) {
		proc.basics[i] = n + i;
		proc.basicflag[n + i] = i;
		proc.x_B[i] = proc.b[i];
	}
	//TODO: lufac was called here
	iterationProcess_dbsolve(&proc, proc.x_B); /* could be done explicitly in terms of bounds/ranges */
	
	/****************************************************************
	 * 	Begin Phase I (i.e., dual simplex method)
	 ****************************************************************/
	
	CHUZC_findAttractiveCandidates(proc.basics, proc.x_B, &attractive);
	if (proc.index < attractive.size) {
		var_out = getACV(attractive, proc.index).var;
	}
	else {
		var_out = -1;
	}
	
	/****************************************************************
	 * 	Main loop                                                   *
	 ****************************************************************/
	for (iter = 0; iter < MAX_ITER && !isProgramFinished(); iter++) {
		print("iteration %d", iter);
//		assert(iter == proc.currentBasis);
		installNewInverse(&proc);
		do {
			keepGoing = 0;
			finished = 0;
			
			// Not sure we need to do basis change and FTRAN inside the loop; it might suffice to do them outside
			applyIncomingBasisChanges(&proc);
			col_out = -1;
			if (var_out != -1 && isBasic(var_out)) {//TODO: isNonbasic in phase II
				col_out = columnForVariable(var_out);
				print("var_out: %d, col_out: %d", var_out, col_out);
				FTRAN_calculateDualStepDirection(&proc, col_out);
			}
			if (var_out == -1 || isNonbasic(var_out) /*TODO: isBasic in phase II*/ || proc.x_B[col_out] > 0 /*TODO: y_N in phase II*/) { //TODO: > EPS or <= -EPS ?
				print("x%d is unattractive: col %d, basicflag %d, stepLength %f", var_out, col_out, (var_out == -1 ? -1 : proc.basicflag[var_out]), (col_out == -1 ? 0 : proc.x_B[col_out]));
				unattractivePayload = (CandidateIsUnattractivePayload*)malloc(sizeof(CandidateIsUnattractivePayload));
				unattractivePayload->var = var_out;
				unattractivePayload->basisNumber = proc.currentBasis;
				send(&proc, "C", "I4->C4", unattractivePayload);
				break; // After the loop, we wait for a new candidate
			}
			else {
				print("x%d is attractive: col %d, basicflag %d, stepLength %f", var_out, col_out, proc.basicflag[var_out], proc.x_B[col_out]);
			}
			chuzrOfferPayload = (ChuzrOfferPayload*)malloc(sizeof(ChuzrOfferPayload));
			chuzrOfferPayload->basisNumber = proc.currentBasis;
			send(&proc, "R", "I5->R1", chuzrOfferPayload);
			while (!(message = receive("R")).sender) {
				if (isProgramFinished()) {
					pthread_exit(NULL);
					return NULL;
				}
				sched_yield();
			}
			assert(streq(message.tag, "R2->I6") || streq(message.tag, "R3->I6"));
			chuzrAcceptancePayload = (ChuzrAcceptancePayload*)message.payload;
			accepted = chuzrAcceptancePayload->accepted;
			free(chuzrAcceptancePayload);
			if (accepted) {
				col_in = CHUZR1_findLeaving_ratioTest(&proc, proc.dyN_dualStepDir, proc.y_N);
				if (col_in == -1) {
					announceProgramFinished();
					print("INFEASIBLE");
					pthread_exit(NULL);
					return NULL;/*TODO: return 2; INFEASIBLE */
				}
				var_in = proc.nonbasics[col_in]; //TODO: basics in PII
				FTRAN_calculatePrimalStepDirection(&proc, col_in);
				CHUZR2_calculateStepLengths(&proc, col_in, col_out, &t_primalStepLength, &s_dualStepLength);
				expandEtaFile(&proc, col_out);
				broadcastBasisChange(&proc, proc.currentBasis, col_in, col_out, var_in, var_out, t_primalStepLength, s_dualStepLength); //TODO: Be careful - this is different in the two phases!
				UPRHS_updatePrimalDualSolutions(&proc, col_in, col_out, t_primalStepLength, s_dualStepLength);
				UPDATE_BASIS(&proc, col_in, col_out, 1);
				++proc.currentBasis;
				
				primal_obj = iterationProcess_recalculateObjectiveValues(proc.c, proc.x_B, proc.basics) + proc.f;
				print("Basis %d:\t%14.7e", iter, high(primal_obj));
				if (proc.currentBasis % 100 == 0) {
					printf("Basis %d:\t%14.7e\n", proc.currentBasis, high(primal_obj));
					fflush(stdout);
				}
				
				CHUZC_findAttractiveCandidates(proc.basics, proc.x_B, &attractive);
				if (attractive.size == 0) {
					finished = 1;
					break; /* ready for Phase II */ //TODO: How to notify the others?
				}
				//col_out = getACV(attractive, 0).col;
				sendAttractiveCandidates(&proc, proc.currentBasis, &attractive);
			}
			else { //Not accepted
				while (!(message = receiveWithTag("I*", "I7->I2I3I10")).sender) {
					if (isProgramFinished()) {
						pthread_exit(NULL);
						return NULL;
					}
					sched_yield();
				}
				assert(streq(message.tag, "I7->I2I3I10"));
				applyIncomingBasisChange(&proc, message);
				keepGoing = 1;
			}
		} while (keepGoing);
		if (finished)
			break;
		
		while (!(message = receive("C")).sender) {
			if (isProgramFinished()) {
				pthread_exit(NULL);
				return NULL;
			}
			sched_yield();
		}
		assert(streq(message.tag, "C3->I11") || streq(message.tag, "C5->I11"));
		applyIncomingBasisChanges(&proc);
		var_out = *(int*)message.payload;
		free(message.payload);
		print("got suggested entering variable: x%d", var_out);
	} /* Phase I loop */
	
	print("End of Phase I");
	if (announceProgramFinished()) {
		pthread_exit(NULL);
		return NULL;
	}
	printf("Phase I solution at basis %d: %14.7e\n", proc.currentBasis, high(primal_obj));
	// At the current stage of development, we are at the point where Phase I 'almost' works (except for a few threading glitches).
	// Therefore, we have been experimenting with continuing with Phaase II, but we are unsure of how all threads should get there properly, not just the thread that discovers that Phase I is over.
	// The above construct will make sure that only one of the threads reaches this point.
	// We have added these two lines in order to shut the program down after Phase I so that we can perform some timing measurements.
	pthread_exit(NULL);
	return NULL;
	
	/****************************************************************
	 *   Restore objective function by setting                       *
	 *                -1  T                                          *
	 *         y  = (B  N) c  - c                                    *
	 *          N           B    N                                   *
	 ****************************************************************/
	
	tempVector.clear();
	for (i = 0; i < m; i++) {
		if (ABS(proc.c[proc.basics[i]]) > ITERATION_EPS) {
			tempVector.append(proc.c[proc.basics[i]], i);
		}
	}
	
	iterationProcess_solveTransposeUsingLU(&proc, &tempVector);

	iterationProcess_Nt_times_y(&proc, proc.AT, proc.basicflag, tempVector, proc.dyN_dualStepDir); // Use dyN_dualStepDir temporarily (this is not the way it is normally used)

	for (j = 0; j < n; j++)
		proc.y_N[j] = -proc.c[proc.nonbasics[j]];
	for (k = 0; k < proc.dyN_dualStepDir.numNonzeroes(); k++) {
		j = proc.dyN_dualStepDir.rowIndex(k);
		proc.y_N[j] += proc.dyN_dualStepDir.value(k);
	}

	/****************************************************************
	 *   Begin Phase II (I.e., primal simplex method)                *
	 ****************************************************************/
	
	for (; iter < MAX_ITER; iter++) {
		primal_obj = iterationProcess_recalculateObjectiveValues(proc.c, proc.x_B, proc.basics) + proc.f;
		print("Iteration %d:\t%14.7e", iter, high(primal_obj));
		
		CHUZC_findAttractiveCandidates(proc.nonbasics, proc.y_N, &attractive);
		if (attractive.size == 0)
			break; /* optimal */
		var_in = getACV(attractive, 0).var;
		col_in = columnForVariable(var_in);
		
		installNewInverse(&proc);
		
		FTRAN_calculatePrimalStepDirection(&proc, col_in);
		
		col_out = CHUZR1_findLeaving_ratioTest(&proc, proc.dxB_primalStepDir, proc.x_B);
		if (col_out == -1) {
			announceProgramFinished();
			print("UNBOUNDED");
			pthread_exit(NULL);
			return NULL; /* return 1; UNBOUNDED */
		}
		var_out = proc.basics[col_out];

		//TODO: Just for testing messaging w/ InvProc
		pivotPayload = (InvProcPivotPayload*)malloc(sizeof(InvProcPivotPayload)); // Will be free'd by InvProc
		pivotPayload->enteringCol = col_in;
		pivotPayload->leavingRow = col_out;
		send(&proc, "V", "I8->V1", pivotPayload);
		
		FTRAN_calculateDualStepDirection(&proc, col_out);
		
		CHUZR2_calculateStepLengths(&proc, col_in, col_out, &t_primalStepLength, &s_dualStepLength);
		
		expandEtaFile(&proc, col_out);
		
		UPRHS_updatePrimalDualSolutions(&proc, col_in, col_out, t_primalStepLength, s_dualStepLength);
		
		UPDATE_BASIS(&proc, col_in, col_out, 1);
		
	} /* End of Phase II */

	primal_obj = iterationProcess_recalculateObjectiveValues(proc.c, proc.x_B, proc.basics) + proc.f;
	print("%8d   %14.7e     NA    ", iter, high(primal_obj));
	print("End of Phase II \n");
	printf("Iterations: %d\nOptimal solution: %14.7e\n", iter, high(primal_obj));

	/****************************************************************
	 * 	Transcribe solution to x vector and dual solution to y      *
	 ****************************************************************/
	
	x = (TYPE*)malloc(N * sizeof(TYPE));
	for (j = 0; j < N; j++)
		x[j] = 0.0;
	for (i = 0; i < m; i++)
		x[proc.basics[i]] = proc.x_B[i];
	//printTypeArray(x, N);
	
	y = (TYPE*)malloc(N * sizeof(TYPE));
	for (j = 0; j < N; j++)
		y[j] = 0.0;
	for (i = 0; i < n; i++)
		y[proc.nonbasics[i]] = proc.y_N[i];
	
	/****************************************************************
	 * 	Split out slack variables and shift dual variables.
	 ****************************************************************/
	
	z = (TYPE*)malloc(n * sizeof(TYPE));
	for (j = 0; j < n; j++)
		z[j] = y[j];
	w = (TYPE*)malloc(m * sizeof(TYPE));
	for (i = 0; i < m; i++) {
		y[i] = y[n + i];
		w[i] = x[n + i];
	}
		
	announceProgramFinished();
	print("OPTIMAL");
	pthread_exit(NULL);
	return NULL;
}

/*-----------------------------------------------------------------+
| Forward/backward solve using LU factorization                    |
| Input:                                                           |
|    m          dimension of array y                               |
|    y          array containing right-hand side                   |
|                                                                  |
|    static global variables (assumed setup by lufac()):           |
|                                                                  |
|    rank       rank of B                                          |
|    L.colPos, L.rowIndices, L, three array sparse representation of L             |
|    UT.colPos,UT.rowIndices,UT.values three array sparse representation of U transpose   |
|               without its diagonal                               |
|    diagU      diagonal entries of U                              |
|    colPerm, invColPerm, rowPerm, invRowPerm                          |
|               column and row permutations and their inverses     |
| Output:                                                          |
|                                          -1                      |
|    y          array containing solution B  y                     |
|                                                                  |
|    integer flag indicating whether system is consistent          |
+-----------------------------------------------------------------*/
// Will update the eta file based on col_out.
static int iterationProcess_solveUsingLU(IterationProcess * proc, SafeSparseVector * y) {
	int i;
	int k, row, consistent = TRUE;
	TYPE beta;
	TYPE eps;
	
	double starttime, endtime;
	
	Tree tree = {NULL, NULL};
	
	starttime = (double) clock();

	for (k = 0; k < y->numNonzeroes(); k++) {
		i = proc->factoredInverse.invRowPerm.at(y->rowIndex(k));
		proc->internalData_SolveUsingLU.yInternal[i] = y->value(k);
		proc->internalData_SolveUsingLU.tagInternal[i] = proc->internalData_SolveUsingLU.currtagInternal;
		addtree(&tree, i);
	}

	if (proc->factoredInverse.rank < y->numRows())
		eps = EPSSOL * y->maxValue();

	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  L  y                                    */

	for (i = getfirst(&tree); i < proc->factoredInverse.rank && i != -1; i = getnext(&tree)) {
		beta = proc->internalData_SolveUsingLU.yInternal[i];
		for (k = proc->factoredInverse.L.colPos[i]; k < proc->factoredInverse.L.colPos[i + 1]; k++) {
			row = proc->factoredInverse.L.rowIndices[k];
			if (proc->internalData_SolveUsingLU.tagInternal[row] != proc->internalData_SolveUsingLU.currtagInternal) {
				proc->internalData_SolveUsingLU.yInternal[row] = 0.0;
				proc->internalData_SolveUsingLU.tagInternal[row] = proc->internalData_SolveUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			proc->internalData_SolveUsingLU.yInternal[row] -= proc->factoredInverse.L.values[k] * beta;
		}
	}

	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  U  y                                    */

	for (i = getlast(&tree); i >= proc->factoredInverse.rank && i != -1; i = getprev(&tree)) {
		if (ABS( proc->internalData_SolveUsingLU.yInternal[i] ) > eps)
			consistent = FALSE;
		proc->internalData_SolveUsingLU.yInternal[i] = 0.0;
	}

	for (; i >= 0; i = getprev(&tree)) {
		beta = proc->internalData_SolveUsingLU.yInternal[i] / proc->factoredInverse.diagU[i];
		for (k = proc->factoredInverse.U.colPos[i]; k < proc->factoredInverse.U.colPos[i + 1]; k++) {
			row = proc->factoredInverse.U.rowIndices[k];
			if (proc->internalData_SolveUsingLU.tagInternal[row] != proc->internalData_SolveUsingLU.currtagInternal) {
				proc->internalData_SolveUsingLU.yInternal[row] = 0.0;
				proc->internalData_SolveUsingLU.tagInternal[row] = proc->internalData_SolveUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			proc->internalData_SolveUsingLU.yInternal[row] -= proc->factoredInverse.U.values[k] * beta;
		}
		proc->internalData_SolveUsingLU.yInternal[i] = beta;
	}

	y->clear();
	for (i = getfirst(&tree); i != -1; i = getnext(&tree)) {
		if (ABS(proc->internalData_SolveUsingLU.yInternal[i]) > SOLVE_USING_LU_EPS) {
			y->append(proc->internalData_SolveUsingLU.yInternal[i], proc->factoredInverse.colPerm.at(i));
		}
	}

	proc->internalData_SolveUsingLU.currtagInternal++;
	killtree(&tree);

	iterationProcess_Gauss_Eta(proc, *y);
	
	/*************************************************************
	 * Update etaFile.values and save col_out in etaFile.newColLocations[etaFile.iter]                   *
	 *************************************************************/
	
	printp("solveUsingLU is modifying eta file");
	for (i = 0, k = proc->etaFile.colPos.at(proc->etaFile.iter); i < y->numNonzeroes(); i++, k++) {
		proc->etaFile.values.push_back(y->value(i));
		proc->etaFile.rowIndices.push_back(y->rowIndex(i));
	}

	proc->etaFile.size = k;
	proc->etaFile.colPos.push_back(k);
	
	endtime = (double) clock();
	proc->cumtime += endtime - starttime;

	return consistent;
}

/*-----------------------------------------------------------------+
 | Forward/backward solve using LU factorization                    |
 | Input:                                                           |
 |    y->rows          dimension of array y                               |
 |    y          array containing right-hand side                   |
 |                                                                  |
 |    static global variables (assumed setup by lufac()):           |
 |                                                                  |
 |    rank       rank of B                                          |
 |    L.colPos, L.rowIndices, L, three array sparse representation of L             |
 |    UT.colPos,UT.rowIndices,UT.values three array sparse representation of U transpose   |
 |               without its diagonal                               |
 |    diagU      diagonal entries of U                              |
 |    colPerm, invColPerm, rowPerm, invRowPerm                          |
 |               column and row permutations and their inverses     |
 | Output:                                                          |
 |                                          -T                      |
 |    y          array containing solution B  y                     |
 |                                                                  |
 |    integer flag indicating whether system is consistent         */

static int iterationProcess_solveTransposeUsingLU(IterationProcess * proc, SafeSparseVector * y) {
	int i;
	int k, row, consistent = TRUE;
	TYPE beta;
	TYPE eps;
	
	double starttime, endtime;
	
	Tree tree = {NULL, NULL};
	
	starttime = (double) clock();

	iterationProcess_Gauss_Eta_T(proc, *y);

	for (k = 0; k < y->numNonzeroes(); k++) {
		i = proc->factoredInverse.invColPerm.at(y->rowIndex(k));
		proc->internalData_SolveTransposeUsingLU.yInternal[i] = y->value(k);
		proc->internalData_SolveTransposeUsingLU.tagInternal[i] = proc->internalData_SolveTransposeUsingLU.currtagInternal;
		addtree(&tree, i);
	}

	if (proc->factoredInverse.rank < y->numRows())
		eps = EPSSOL * y->maxValue();

	/*------------------------------------------------------+
	 |               -T                                      |
	 |       y  <-  U  y                                    */

	for (i = getfirst(&tree); i < proc->factoredInverse.rank && i != -1; i = getnext(&tree)) {
		beta = proc->internalData_SolveTransposeUsingLU.yInternal[i] / proc->factoredInverse.diagU[i];
		for (k = proc->factoredInverse.UT.colPos[i]; k < proc->factoredInverse.UT.colPos[i + 1]; k++) {
			row = proc->factoredInverse.UT.rowIndices[k];
			if (proc->internalData_SolveTransposeUsingLU.tagInternal[row] != proc->internalData_SolveTransposeUsingLU.currtagInternal) {
				proc->internalData_SolveTransposeUsingLU.yInternal[row] = 0.0;
				proc->internalData_SolveTransposeUsingLU.tagInternal[row] = proc->internalData_SolveTransposeUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			proc->internalData_SolveTransposeUsingLU.yInternal[row] -= proc->factoredInverse.UT.values[k] * beta;
		}
		proc->internalData_SolveTransposeUsingLU.yInternal[i] = beta;
	}
	for (i = getlast(&tree); i >= proc->factoredInverse.rank && i != -1; i = getprev(&tree)) {
		if (ABS( proc->internalData_SolveTransposeUsingLU.yInternal[i] ) > eps)
			consistent = FALSE;
		proc->internalData_SolveTransposeUsingLU.yInternal[i] = 0.0;
	}

	/*------------------------------------------------------+
	 |               -T                                      |
	 |       y  <-  L  y                                    */

	for (; i >= 0; i = getprev(&tree)) {
		beta = proc->internalData_SolveTransposeUsingLU.yInternal[i];
		for (k = proc->factoredInverse.LT.colPos[i]; k < proc->factoredInverse.LT.colPos[i + 1]; k++) {
			row = proc->factoredInverse.LT.rowIndices[k];
			if (proc->internalData_SolveTransposeUsingLU.tagInternal[row] != proc->internalData_SolveTransposeUsingLU.currtagInternal) {
				proc->internalData_SolveTransposeUsingLU.yInternal[row] = 0.0;
				proc->internalData_SolveTransposeUsingLU.tagInternal[row] = proc->internalData_SolveTransposeUsingLU.currtagInternal;
				addtree(&tree, row);
			}
			proc->internalData_SolveTransposeUsingLU.yInternal[row] -= proc->factoredInverse.LT.values[k] * beta;
		}
	}

	y->clear();
	for (i = getfirst(&tree); i != -1; i = getnext(&tree)) {
		if (ABS(proc->internalData_SolveTransposeUsingLU.yInternal[i]) > SOLVE_USING_LU_EPS) {
			y->append(proc->internalData_SolveTransposeUsingLU.yInternal[i], proc->factoredInverse.rowPerm.at(i));
		}
	}

	proc->internalData_SolveTransposeUsingLU.currtagInternal++;
	killtree(&tree);

	endtime = (double) clock();
	proc->cumtime += endtime - starttime;

	return consistent;
}

/*
static void iterationProcess_lu_clo(IterationProcess * proc) {
	freeIV( &proc->factoredInverse.rowPerm );
	freeIV( &proc->factoredInverse.invRowPerm );
	freeIV( &proc->factoredInverse.colPerm );
	freeIV( &proc->factoredInverse.invColPerm );
	FREE( proc->factoredInverse.L.values );
	FREE( proc->factoredInverse.L.rowIndices );
	FREE( proc->factoredInverse.L.colPos );
	FREE( proc->factoredInverse.UT.values );
	FREE( proc->factoredInverse.UT.rowIndices );
	FREE( proc->factoredInverse.UT.colPos );
	FREE( proc->factoredInverse.diagU );
	freeIV( &proc->etaFile.newColLocations );
	freeTV( &proc->etaFile.values );
	freeIV( &proc->etaFile.rowIndices );
	freeIV( &proc->etaFile.colPos );
}
*/

/*-----------------------------------------------------------------+
 | Forward/backward solve using LU factorization                    |
 | Input:                                                           |
 |    m          dimension of array y                               |
 |    y          array containing right-hand side                   |
 |                                                                  |
 |    static global variables (assumed setup by lufac()):           |
 |                                                                  |
 |    rank       rank of B                                          |
 |    L.colPos, L.rowIndices, L.values, three array sparse representation of L.values             |
 |    UT.colPos,UT.rowIndices,UT.values three array sparse representation of U transpose   |
 |               without its diagonal                               |
 |    diagU      diagonal entries of U                              |
 |    colPerm, invColPerm, rowPerm, invRowPerm                          |
 |               column and row permutations and their inverses     |
 | Output:                                                          |
 |                                          -1                      |
 |    y          array containing solution B  y                     |
 |                                                                  |
 |    integer flag indicating whether system is consistent         */

static int iterationProcess_dbsolve(IterationProcess * proc, SafeVector<TYPE> & y) {
	int i;
	int k, row, consistent = TRUE;
	TYPE beta, *dwork;
	TYPE eps;
	int m = y.size();
	
	double starttime, endtime;

	starttime = (double) clock();

	MALLOC (dwork,m,TYPE);

	if (proc->factoredInverse.rank < m)
		eps = EPSSOL * maxv(y);
	for (i = 0; i < m; i++)
		dwork[i] = y[i];
	for (i = 0; i < m; i++)
		y[proc->factoredInverse.invRowPerm.at(i)] = dwork[i];
	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  L  y                                    */

	for (i = 0; i < proc->factoredInverse.rank; i++) {
		beta = y[i];
		for (k = proc->factoredInverse.L.colPos[i]; k < proc->factoredInverse.L.colPos[i + 1]; k++) {
			row = proc->factoredInverse.L.rowIndices[k];
			y[row] -= proc->factoredInverse.L.values[k] * beta;
		}
	}

	/*------------------------------------------------------+
	 |               -1                                      |
	 |       y  <-  U  y                                    */

	for (i = m - 1; i >= proc->factoredInverse.rank; i--) {
		if (ABS( y[i] ) > eps)
			consistent = FALSE;
		y[i] = 0.0;
	}

	for (i = proc->factoredInverse.rank - 1; i >= 0; i--) {
		beta = y[i];
		for (k = proc->factoredInverse.UT.colPos[i]; k < proc->factoredInverse.UT.colPos[i + 1]; k++) {
			beta -= proc->factoredInverse.UT.values[k] * y[proc->factoredInverse.UT.rowIndices[k]];
		}
		y[i] = beta / proc->factoredInverse.diagU[i];
	}

	for (i = 0; i < m; i++)
		dwork[i] = y[i];
	for (i = 0; i < m; i++)
		y[proc->factoredInverse.colPerm.at(i)] = dwork[i];

	FREE(dwork);

	endtime = (double) clock();
	proc->cumtime += endtime - starttime;

	return consistent;
}

static void iterationProcess_Gauss_Eta(IterationProcess * proc, SafeSparseVector & dx_B) {
	int i, j, k, col, kcol, ii;
	TYPE temp;
	
	if (proc->etaFile.iter <= 0) {
		printp("Gauss_Eta is returning due to empty eta file");
		return;
	}

	ii = -1;
	for (k = 0; k < dx_B.numNonzeroes(); k++) {
		i = dx_B.rowIndex(k);
		proc->internalData_Gauss_Eta.a[i] = dx_B.value(k);
		proc->internalData_Gauss_Eta.tag[i] = proc->internalData_Gauss_Eta.currtag;
		proc->internalData_Gauss_Eta.link[ii] = i;
		ii = i;
	}
	printp("Gauss_Eta is reading eta file");
	for (j = 0; j < proc->etaFile.iter; j++) {
		assert(proc->etaFile.colPos.at(j) < proc->etaFile.colPos.at(j + 1));
		col = proc->etaFile.newColLocations.at(j);

		for (k = proc->etaFile.colPos.at(j); k < proc->etaFile.colPos.at(j + 1); k++) {
			i = proc->etaFile.rowIndices.at(k);
			if (proc->internalData_Gauss_Eta.tag[i] != proc->internalData_Gauss_Eta.currtag) {
				proc->internalData_Gauss_Eta.a[i] = 0.0;
				proc->internalData_Gauss_Eta.tag[i] = proc->internalData_Gauss_Eta.currtag;
				proc->internalData_Gauss_Eta.link[ii] = i;
				ii = i;
			}
			if (i == col)
				kcol = k;
		}

		temp = proc->internalData_Gauss_Eta.a[col] / proc->etaFile.values.at(kcol);
		if (temp != 0.0) {
			for (k = proc->etaFile.colPos.at(j); k < kcol; k++) {
				i = proc->etaFile.rowIndices.at(k);
				proc->internalData_Gauss_Eta.a[i] -= proc->etaFile.values.at(k) * temp;
			}
			proc->internalData_Gauss_Eta.a[col] = temp;
			for (k = kcol + 1; k < proc->etaFile.colPos.at(j + 1); k++) {
				i = proc->etaFile.rowIndices.at(k);
				proc->internalData_Gauss_Eta.a[i] -= proc->etaFile.values.at(k) * temp;
			}
		}
	}
	proc->internalData_Gauss_Eta.link[ii] = dx_B.numRows();
	proc->internalData_Gauss_Eta.currtag++;
	
	dx_B.clear();
	for (i = proc->internalData_Gauss_Eta.link[-1]; i < dx_B.numRows(); i = proc->internalData_Gauss_Eta.link[i]) {
		if (ABS(proc->internalData_Gauss_Eta.a[i]) > GAUSS_ETA_EPS) {
			dx_B.append(proc->internalData_Gauss_Eta.a[i], i);
		}
	}
}

/*****************************************************************
 *  Gaussian elimination for Eta transformations which will solve *
 * each system B y = c for y                             	 *
 *****************************************************************/

static void iterationProcess_Gauss_Eta_T(IterationProcess * proc, SafeSparseVector & vec) {
	int i = -1, j, k, kk = -1, kkk = -1, col;
	TYPE temp;
	
	//printp("Gauss_Eta_T is reading eta file");
	for (j = proc->etaFile.iter - 1; j >= 0; j--) {
		col = proc->etaFile.newColLocations.at(j);

		for (k = 0; k < vec.numNonzeroes(); k++) {
			i = vec.rowIndex(k);
			if (i == col)
				kk = k;
			proc->internalData_Gauss_Eta_T.a[i] = vec.value(k);
			proc->internalData_Gauss_Eta_T.tag[i] = proc->internalData_Gauss_Eta_T.currtag;
		}

		if (proc->internalData_Gauss_Eta_T.tag[col] != proc->internalData_Gauss_Eta_T.currtag) {
			kk = vec.numNonzeroes();
			vec.append(0.0, col);
			proc->internalData_Gauss_Eta_T.a[col] = 0.0;
			proc->internalData_Gauss_Eta_T.tag[col] = proc->internalData_Gauss_Eta_T.currtag;
		}
		temp = vec.value(kk);
		for (k = proc->etaFile.colPos.at(j); k < proc->etaFile.colPos.at(j + 1); k++) {
			i = proc->etaFile.rowIndices.at(k);
			if (i == col)
				kkk = k;
			if (proc->internalData_Gauss_Eta_T.tag[i] == proc->internalData_Gauss_Eta_T.currtag) {
				if (i != col) {
					temp -= proc->etaFile.values.at(k) * proc->internalData_Gauss_Eta_T.a[i];
				}
			}
		}
		proc->internalData_Gauss_Eta_T.currtag++;

		vec.value(kk) = temp / proc->etaFile.values.at(kkk);
	}
}

static void iterationProcess_Nt_times_y(IterationProcess * proc, SparseMatrix AT, const SafeVector<int> & basicflag, const SafeSparseVector & y, SafeSparseVector & yN) {
	int i, j, jj, k, kk;
	
	jj = -1;
	for (k = 0; k < y.numNonzeroes(); k++) {
		i = y.rowIndex(k);
		for (kk = AT.colPos[i]; kk < AT.colPos[i + 1]; kk++) {
			j = AT.rowIndices[kk];
			if (basicflag.at(j) < 0) {
				if (proc->internalData_Nt_times_y.tag[j] != proc->internalData_Nt_times_y.currtag) {
					proc->internalData_Nt_times_y.a[j] = 0.0;
					proc->internalData_Nt_times_y.tag[j] = proc->internalData_Nt_times_y.currtag;
					proc->internalData_Nt_times_y.link[jj] = j;
					jj = j;
				}
				proc->internalData_Nt_times_y.a[j] += y.value(k) * AT.values[kk];
			}
		}
	}
	proc->internalData_Nt_times_y.link[jj] = AT.rows;
	proc->internalData_Nt_times_y.currtag++;

	yN.clear();
	for (jj = proc->internalData_Nt_times_y.link[-1]; jj < AT.rows; jj = proc->internalData_Nt_times_y.link[jj]) {
		if (ABS(proc->internalData_Nt_times_y.a[jj]) > EPS_MATRIX) {
			yN.append(proc->internalData_Nt_times_y.a[jj], columnForVariableP(jj));
		}
	}
}

/*************************************************************
 * STEP 3: Ratio test to find leaving column                  *
 *************************************************************/
static int CHUZR1_findLeaving_ratioTest(IterationProcess * proc, const SafeSparseVector & dy, const SafeVector<TYPE> & y) {
	int j, jj = -1, k;
	TYPE min = HUGE_VAL;
	//printp("ratio test");
	//printRawSparseVector(dy);
	for (k = 0; k < dy.numNonzeroes(); k++) {
		if (dy.value(k) > RATIO_TEST_ETA) {
			j = dy.rowIndex(k);
			if (y[j] / dy.value(k) < min) {
				min = y[j] / dy.value(k);
				jj = j;
			}
		}
	}
	return jj;
}

static TYPE iterationProcess_recalculateObjectiveValues(const SafeVector<TYPE> & c, const SafeVector<TYPE> & x_B, const SafeVector<int> & basics) {
	int i;
	TYPE prod = 0.0;

	for (i = 0; i < x_B.size(); i++) {
		prod += c[basics[i]] * x_B[i];
	}

	return prod;
}

static int compareAttractiveCandidates(const void * a, const void * b) {
	const AttractiveCandidate * candA = (const AttractiveCandidate *)a;
	const AttractiveCandidate * candB = (const AttractiveCandidate *)b;
	if (candA->value < candB->value)
		return -1;
	else if (candA->value > candB->value)
		return 1;
	else
		return 0;
}

/************************************************************************************
 * STEP 1: Pick most negative nonbasic (primal or dual, depending on the parameters) *
 *         'attractive' is assumed to be initialised. 
 ************************************************************************************/
static void CHUZC_findAttractiveCandidates(const SafeVector<int> & columnToVariableMapping, const SafeVector<TYPE> & x, AttractiveCandidateVector * attractive) {
	int i;
	TYPE best = x[0];
	attractive->size = 0;
	for (i = 0; i < x.size(); i++) {
		if (x[i] < best) {
			best = x[i];
		}
		if (x[i] < -PICK_NEG_EPS) {
			AttractiveCandidate cand;
			cand.var = columnToVariableMapping[i];
			cand.value = x[i];
			appendACV(attractive, cand);
		}
	}
	//printf("best candidate: %e\n", best);fflush(stdout);
	qsort(attractive->array, attractive->size, sizeof(AttractiveCandidate), compareAttractiveCandidates);
}
