#include "Matrix.h"
#define VECTOR_WIDTH (16 / sizeof(TYPE))
#define ALIGN_LOG2 4
#define ROUND_UP_MULTIPLE(x, m) (((x) + (m) - 1) / m * m) // Returns x rounded up to the nearest multiple of m

#ifdef __powerpc__
	// On Cell, we need these includes
	// Note: __powerpc__ should perhaps be replaced by something more Cell-specific in case one wants to compile this on a PowerPC that is not a Cell PPE
	#include <altivec.h>
	#include <libmisc.h>
#else
	#include <cstdlib>
	// On x86, there is no malloc_align and no need for it either
	#define malloc_align(size, alignment) malloc(size)
	#define free_align(buffer) free(buffer)
#endif

#ifdef USE_SIMD
	#ifdef USE_GMP
		#error "Cannot use GMP with SIMD operations"
	#endif
#endif

using namespace std;

Matrix::Matrix(int rows, int cols) {
	this->rows = rows;
	this->cols = cols;
#ifdef USE_GMP
	// GMP's types are classes and require their constructors to be called, so we need to use new.
	// If necessary, it is possible to combine malloc_align and GMP through the use of 'placement new'.
	this->physicalCols = cols;
	this->data = new mpq_class[rows * this->physicalCols];
#else
	this->physicalCols = ROUND_UP_MULTIPLE(cols, VECTOR_WIDTH);
	this->data = (TYPE*)malloc_align(rows * this->physicalCols * sizeof(TYPE), ALIGN_LOG2);
#endif
	for (int i = 0; i < rows * this->physicalCols; ++i)
		this->data[i] = 0;
}

Matrix::Matrix(int rows, int cols, TYPE * data) {
	this->rows = rows;
	this->cols = cols;
#ifdef USE_GMP
	this->physicalCols = cols;
	this->data = new TYPE[rows * this->physicalCols];
#else
	this->physicalCols = ROUND_UP_MULTIPLE(cols, VECTOR_WIDTH);
	this->data = (TYPE*)malloc_align(rows * this->physicalCols * sizeof(TYPE), ALIGN_LOG2);
#endif
	for (int r = 0; r < rows; ++r) {
		for (int c = 0; c < cols; ++c)
			this->data[r * this->physicalCols + c] = data[r * cols + c];
		for (int c = cols; c < this->physicalCols; ++c)
			this->data[r * this->physicalCols + c] = 0;
	}
}

Matrix::Matrix(int rows, int cols, bool identity) {
	this->rows = rows;
	this->cols = cols;
#ifdef USE_GMP
	this->physicalCols = cols;
	this->data = new TYPE[rows * this->physicalCols];
#else
	this->physicalCols = ROUND_UP_MULTIPLE(cols, VECTOR_WIDTH);
	this->data = (TYPE*)malloc_align(rows * this->physicalCols * sizeof(TYPE), ALIGN_LOG2);
#endif
	for (int i = 0; i < rows * this->physicalCols; ++i)
		this->data[i] = 0;
	if (identity && rows == cols) {
		for (int i = 0; i < rows; ++i) {
			(*this)(i, i) = 1;
		}
	}
}

Matrix::Matrix(const Matrix & source) {
	this->rows = source.rows;
	this->cols = source.cols;
	this->physicalCols = source.physicalCols;
#ifdef USE_GMP
	this->data = new TYPE[source.rows * source.physicalCols];
#else
	this->data = (TYPE*)malloc_align(source.rows * source.physicalCols * sizeof(TYPE), ALIGN_LOG2);
#endif
	for (int i = 0; i < source.rows * source.physicalCols; ++i)
		this->data[i] = source.data[i];
}

Matrix::~Matrix() {
#ifdef USE_GMP
	delete [] data;
#else
	free_align((float*)data);
#endif
}

// Prints the entire matrix
ostream & operator << (ostream & out, const Matrix & matrix) {
	out << "=== " << matrix.rows << " x " << matrix.cols << " @ " << matrix.data << " ===" << endl;
	for (int r = 0; r < matrix.rows; ++r) {
		out << matrix(r, 0);
		for (int c = 1; c < matrix.cols; ++c)
			out << ' ' << matrix(r, c);
		out << endl;
	}
	out << "======" << endl;
	return out;
}

// A more advanced print operation that prints only nonzero entries, along with variable names
void Matrix::print(const vector<int> & basic, const vector<int> & nonbasic) {
	cout << "=== " << rows << " x " << cols << " @ " << data << " ===" << endl;
	for (int r = 0; r < rows; ++r) {
		if (r == 0)
			cout << "z = ";
		else
			cout << "x" << basic[r - 1] << " = ";
		cout << (*this)(r, 0);
		for (int c = 1; c < cols; ++c)
			if ((*this)(r, c) != 0)
				cout << " " << (*this)(r, c) << "x" << nonbasic[c - 1];
		cout << endl;
	}
	cout << "======" << endl;
}

// No point in using SIMD here; this function is used very rarely
void Matrix::swapRows(int firstRow, int secondRow) {
	if (firstRow == secondRow) return;
	for (int j = 0; j < cols; ++j) {
		TYPE tmp = (*this)(firstRow, j);
		(*this)(firstRow, j) = (*this)(secondRow, j);
		(*this)(secondRow, j) = tmp;
	}
}



#ifdef USE_SIMD

	void Matrix::multiplyRow(int row, TYPE factor) {
		if (factor == 1) return;
		vector TYPE zero_v = (vector TYPE){0.0f, 0.0f, 0.0f, 0.0f};
		vector TYPE factor_v = (vector TYPE){factor, factor, factor, factor}; // Wanted to use vec_splat(vec_lde(0, &factor), 0) here, but might have misunderstood the syntax
		vector TYPE * data_v = (vector TYPE *)(data + row * physicalCols);
		for (int j = 0; j < physicalCols / VECTOR_WIDTH; ++j) {
			data_v[j] = vec_madd(data_v[j], factor_v, zero_v);
		}
	}

	void Matrix::addRows(int sourceRow, int destinationRow, TYPE factor) {
		if (factor == 0) return;
		vector TYPE factor_v = (vector TYPE){factor, factor, factor, factor}; // As above
		vector TYPE * source_v = (vector TYPE *)(data + sourceRow * physicalCols);
		vector TYPE * destination_v = (vector TYPE *)(data + destinationRow * physicalCols);
		for (int j = 0; j < physicalCols / VECTOR_WIDTH; ++j) {
			destination_v[j] = vec_madd(source_v[j], factor_v, destination_v[j]);
		}
	}

#else

	void Matrix::multiplyRow(int row, TYPE factor) {
		if (factor == 1) return;
		for (int j = 0; j < cols; ++j) {
			(*this)(row, j) *= factor;
		}
	}

	void Matrix::addRows(int sourceRow, int destinationRow, TYPE factor) {
		if (factor == 0) return;
		for (int j = 0; j < cols; ++j) {
			incr((*this)(destinationRow, j), (*this)(sourceRow, j) * factor);
		}
	}

#endif

