#define PPE_CODE

#include "Matrix.h"
#include "types.h"
#include "timer.h"
#include "mps.h"
#include <libspe2.h>
#include <libmisc.h>
#include <pthread.h>
#include <iostream>
#include <fstream>
using namespace std;

#define MAX_SPE_THREADS 8

extern spe_program_handle_t speProgramHandle; // Must match the name given in the SPE makefile

struct PpuThreadData {
	spe_context_ptr_t speContext;
	pthread_t pthread;
	void * argument;
};

SimplexResult simplexIteration(vector<int> & basic, vector<int> & nonbasic, int numSpeThreads, spe_context_ptr_t speContexts[]);

void * ppuPthreadFunction(void * argument) {
	PpuThreadData * data = (PpuThreadData *)argument;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	if (spe_context_run(data->speContext, &entry, 0, data->argument, NULL, NULL) < 0) {
		perror("Failed running context");
		exit(1);
	}
	pthread_exit(NULL);
}

int main(int argc, char * argv[]) {
	int numSpeThreads;
	PpuThreadData data[MAX_SPE_THREADS];
	ParameterContext contexts[MAX_SPE_THREADS] ALIGNED_QUAD;
	spe_context_ptr_t speContexts[MAX_SPE_THREADS];
	
	if (argc < 2) {
		cout << "Must take MPS file name as parameter; the second argument is optional and specifies the number of SPEs" << endl;
		return 1;
	}
	ifstream infile(argv[1]);
	vector<vector<float> > parsedTableau = parse(infile);
	int rows = parsedTableau.size();
	int cols = parsedTableau[0].size();
	Matrix matrix(rows, cols);
	for (int r = 0; r < rows; ++r) {
		for (int c = 0; c < cols; ++c) {
			matrix(r, c) = parsedTableau[r][c];
		}
		if (r > 0 && matrix(r, 0) < 0) {
			cout << "This prototype does only handle Phase II; must be run on problems that are initially feasible (the entire right hand side must be nonnegative)" << endl;
			return 1;
		}
	}
	volatile float * pivotRow = (volatile float *)malloc_align(matrix.getPhysicalCols() * sizeof(float), ALIGN_QUAD_LOG2);
	
	unsigned long long startTime = rdtsc();
	
	int n = cols - 1;
	int m = rows - 1;
	vector<int> basic, nonbasic;
	// Nonbasic variables are labeled 1 .. n
	for (int i = 1; i <= n; ++i)
		nonbasic.push_back(i);
	// Basic variables are labeled n+1 .. n+m
	for (int i = n + 1; i <= n + m; ++i)
		basic.push_back(i);
	
	// Determine how many SPE threads we want
	if (argc >= 3) {
		numSpeThreads = atoi(argv[2]);
	}
	else {
		numSpeThreads = spe_cpu_info_get(SPE_COUNT_USABLE_SPES, -1);
	}
	if (numSpeThreads < 1)
		numSpeThreads = 1;
	if (numSpeThreads > MAX_SPE_THREADS)
		numSpeThreads = MAX_SPE_THREADS;
	if (numSpeThreads > matrix.getRows())
		numSpeThreads = matrix.getRows();
	cout << "Using " << numSpeThreads << " SPEs" << endl;
	
	// Create and start SPE threads
	for (int i = 0; i < numSpeThreads; ++i) {
		contexts[i].rows = matrix.getRows();
		contexts[i].cols = matrix.getCols();
		contexts[i].physicalCols = matrix.getPhysicalCols();
		contexts[i].startingRow = matrix.getRows() / numSpeThreads * i;
		if (i == numSpeThreads - 1)
			contexts[i].numContainedRows = matrix.getRows() - contexts[i].startingRow;
		else
			contexts[i].numContainedRows = matrix.getRows() / numSpeThreads;
		contexts[i].spuIndex = i;
		contexts[i].dataOrigin = (float *)matrix.getData(); // Removing 'volatile'ness is ok here, since this pointer will just be used by the SPU to indicate the MFC target
		contexts[i].ppuPivotRow = (float *)pivotRow;
		
		if ((data[i].speContext = spe_context_create(0, NULL)) == NULL) {
			perror("Failed creating context");
			exit(1);
		}
		if (spe_program_load(data[i].speContext, &speProgramHandle)) {
			perror("Failed loading program");
			exit(1);
		}
		speContexts[i] = data[i].speContext;
		data[i].argument = &contexts[i];
		if (pthread_create(&data[i].pthread, NULL, &ppuPthreadFunction, &data[i])) {
			perror("Failed creating thread");
			return 1;
		}
	}
	
	int numIterations = 0;
	while (simplexIteration(basic, nonbasic, numSpeThreads, speContexts) == SUBOPTIMAL) {
	//	cout << "Done with iteration " << ++numIterations << endl;
	}
	
	// Uncomment this if you want the list of basic/nonbasic variables to be printed
	/*
	cout << "Basic:";
	for (uint i = 0; i < basic.size(); ++i)
		cout << " " << basic[i];
	cout << "\nNonbasic:";
	for (uint i = 0; i < nonbasic.size(); ++i)
		cout << " " << nonbasic[i];
	cout << endl;
	*/
	
	// Wait for SPEs to terminate
	for (int i = 0; i < numSpeThreads; ++i) {
		if (pthread_join(data[i].pthread, NULL)) {
			perror("Failed joining thread");
			exit(1);
		}
		if (spe_context_destroy(data[i].speContext) != 0) {
			perror("Failed destroying context");
			exit(1);
		}
	}
	
	cout << "The optimal solution is " << matrix(0, 0) << endl;
	cout << "Elapsed time (minus input parsing): " << rdtsc() - startTime << endl;
	return 0;
}



// For explanation of communication tags, see spu/SpuTableauSimplex.cpp
SimplexResult simplexIteration(vector<int> & basic, vector<int> & nonbasic, int numSpeThreads, spe_context_ptr_t speContexts[]) {
	//comm0
	while (!spe_out_mbox_status(speContexts[0]));
	uint entering;
	spe_out_mbox_read(speContexts[0], &entering, 1);
	
	//comm1
	for (int i = 1; i < numSpeThreads; ++i) {
		spe_in_mbox_write(speContexts[i], &entering, 1, SPE_MBOX_ALL_BLOCKING); // entering may be SIMPLEX_MBOX_OPTIMAL, in which case SPE code will terminate
	}
	if (entering == SIMPLEX_MBOX_OPTIMAL) {
		return OPTIMAL;
	}
	
	//comm2
	// The case where the first SPE only contains the objective function will not cause problems. That SPE will report SIMPLEX_MBOX_UNBOUNDED since it doesn't find a leaving variable, so it will be ignored unless no other SPEs find a leaving variable - in which case the entire problem is unbounded.
	while (!spe_out_mbox_status(speContexts[0]));
	float maxLeavingValue = 0; // Initialise with arbitrary value to get rid of warning
	int maxLeavingSpu = -1;
	int maxLeavingIndex = -1;
	for (int i = 0; i < numSpeThreads; ++i) {
		Value32 leavingValue, leavingIndex;
		while (!spe_out_mbox_status(speContexts[i]));
		spe_out_mbox_read(speContexts[i], &leavingValue.uintValue, 1);
		while (!spe_out_mbox_status(speContexts[i]));
		spe_out_mbox_read(speContexts[i], &leavingIndex.uintValue, 1);
		//cout << i << " reports " << leavingValue.floatValue << endl;
		if (leavingValue.uintValue != SIMPLEX_MBOX_UNBOUNDED && (maxLeavingSpu == -1 || leavingValue.floatValue > maxLeavingValue)) {
			maxLeavingValue = leavingValue.floatValue;
			maxLeavingIndex = leavingIndex.intValue;
			maxLeavingSpu = i;
		}
	}
	
	//comm3
	for (int i = 0; i < numSpeThreads; ++i) {
		uint instruction;
		if (maxLeavingSpu == -1)
			instruction = SIMPLEX_MBOX_UNBOUNDED;
		else if (maxLeavingSpu == i)
			instruction = SIMPLEX_MBOX_LEAVING_IS_HERE;
		else
			instruction = SIMPLEX_MBOX_LEAVING_IS_ELSEWHERE;
		spe_in_mbox_write(speContexts[i], &instruction, 1, SPE_MBOX_ALL_BLOCKING);
	}
	if (maxLeavingSpu == -1) {
		cout << "Unbounded" << endl;
		return UNBOUNDED; //TODO: cleanup
	}

//	cout << "Pivoting: variable in row " << maxLeavingIndex << " (in SPU " << maxLeavingSpu << ") leaves, variable in column " << entering << " enters" << endl;
	
	//comm4 - sync
	uint dummySyncValue;
	while (!spe_out_mbox_status(speContexts[maxLeavingSpu]));
	spe_out_mbox_read(speContexts[maxLeavingSpu], &dummySyncValue, 1);
	
	//comm5
	for (int i = 0; i < numSpeThreads; ++i) {
		if (maxLeavingSpu != i) {
			spe_in_mbox_write(speContexts[i], &dummySyncValue, 1, SPE_MBOX_ALL_BLOCKING);
		}
	}
	
	// Swap basic and nonbasic variables
	int tmp = basic[maxLeavingIndex - 1];
	basic[maxLeavingIndex - 1] = nonbasic[entering - 1];
	nonbasic[entering - 1] = tmp;
	
	return SUBOPTIMAL;
}
