#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <vector>
#include <stdexcept>
#include <sstream>
#include <cctype>
#include "gmpInterop.h"

using namespace std;

enum RowType {
	LESS_THAN, 
	EQUAL_TO, 
	GREATER_THAN, 
	OBJECTIVE
};

string rowTypeLabels[] = {"L", "E", "G", "N"};

class Equation {
public:
	string label;
	RowType type;
	map<string, TYPE> values;
	int index;
	TYPE rhs;
	
	Equation(string label, string type, int index) {
		this->label = label;
		if (type == "L")
			this->type = LESS_THAN;
		else if (type == "E")
			this->type = EQUAL_TO;
		else if (type == "G")
			this->type = GREATER_THAN;
		else if (type == "N")
			this->type = OBJECTIVE;
		else
			throw invalid_argument("");
		this->index = index;
		this->rhs = 0;
	}
};

string strip(string line) {
	int start = 0, end = (int)line.size() - 1;
	while (start < (int)line.size() && isspace(line[start])) ++start;
	while (end >= 0 && isspace(line[end])) --end;
	if (end < start)
		return "";
	else
		return line.substr(start, end - start + 1);
}

vector<string> split(string line) {
	stringstream ss(line);
	vector<string> items;
	string item;
	while (ss >> item) {
		items.push_back(item);
	}
	return items;
}

vector<vector<TYPE> > parse(const vector<string> & lines) {
	map<string, Equation *> equations;
	vector<string> columnLabels;
	map<string, int> columnIndices;
	vector<vector<TYPE> > tableau;
	unsigned int i = 0;
	int objectiveIndex = -1;
	while (i < lines.size()) {
		string line = lines[i];
		string header = strip(line);
		i++;
		if (line[0] == '*' || line[0] == ' ' || line.substr(0, 4) == "NAME") {
			continue;
		}
		else if (strip(line) == "ENDATA") {
			break;
		}
		else if (strip(line) == "ROWS") {
			int rowIndex = 0;
			while (lines[i][0] == ' ') {
				vector<string> items = split(lines[i]);
				Equation * eqn = new Equation(strip(items[1]), strip(items[0]), rowIndex);
				if (eqn->type == OBJECTIVE)
					objectiveIndex = rowIndex;
				equations[eqn->label] = eqn;
				rowIndex++;
				i++;
			}
		}
		else if (strip(line) == "COLUMNS") {
			int columnIndex = -1;
			while (lines[i][0] == ' ') {
				vector<string> items = split(lines[i]);
				int lim = (items.size() == 5 ? 2 : 1);
				string columnLabel = strip(items[0]);
				if (columnIndices.find(columnLabel) == columnIndices.end()) {
					columnIndex++;
					columnLabels.push_back(columnLabel);
					columnIndices[columnLabel] = columnIndex;
				}
				for (int j = 0; j < lim; ++j) {
					string rowLabel = strip(items[1 + j * 2]);
					stringstream ss(strip(items[2 + j * 2]));
					TYPE value;
					readNumber(ss, value);
					equations[rowLabel]->values[columnLabel] = value;
				}
				i++;
			}
		}
		else if (strip(line) == "RHS") {
			while (lines[i][0] == ' ') {
				vector<string> items = split(lines[i]);
				int lim = (items.size() == 5 ? 2 : 1);
				for (int j = 0; j < lim; ++j) {
					string rowLabel = strip(items[1 + j * 2]);
					stringstream ss(strip(items[2 + j * 2]));
					TYPE value;
					readNumber(ss, value);
					equations[rowLabel]->rhs = value;
				}
				i++;
			}
		}
		else {
			throw invalid_argument("");
		}
	}
	tableau.resize(1); // obj. func.
	for (map<string, Equation *>::iterator eqnIter = equations.begin(); eqnIter != equations.end(); ++eqnIter) {
		Equation * eqn = eqnIter->second;
		vector<TYPE> row(columnLabels.size() + 1, 0);
		bool any = false;
		row[0] = eqn->rhs;
		for (map<string, TYPE>::iterator colIter = eqn->values.begin(); colIter != eqn->values.end(); ++colIter) {
			string colLabel = colIter->first;
			row[columnIndices[colLabel] + 1] = -eqn->values[colLabel];
			if (eqn->values[colLabel] != 0)
				any = true;
		}
		if (!any)
			continue;
		if (eqn->type == OBJECTIVE) {
			tableau[0] = row;
		}
		else {
			if (eqn->type == LESS_THAN || eqn->type == EQUAL_TO) {
				tableau.push_back(row);
			}
			if (eqn->type == GREATER_THAN || eqn->type == EQUAL_TO) {
				for (unsigned int j = 0; j < row.size(); ++j)
					row[j] = -row[j];
				tableau.push_back(row);
			}
		}
		delete eqn;
	}
	
	return tableau;
}

vector<vector<TYPE> > parse(istream & input) {
	string line;
	vector<string> lines;
	while (getline(input, line)) {
		lines.push_back(line);
	}
	return parse(lines);
}
