#include "PartialMatrix.h"
#include <spu_intrinsics.h>
#include <libmisc.h>
#include "../types.h"
#include <stdio.h>

using namespace std;

void PartialMatrix::init(int rows, int cols, int physicalCols, int startRow, int containedRows) {
	this->rows = rows;
	this->cols = cols;
	this->physicalCols = ROUND_UP_MULTIPLE(cols, VECTOR_WIDTH);
	this->startRow = startRow;
	this->containedRows = containedRows;
	this->data = (float*)malloc_align(containedRows * physicalCols * sizeof(float), ALIGN_QUAD_LOG2);
}

PartialMatrix::PartialMatrix(int rows, int cols, int physicalCols, int startRow, int containedRows) {
	init(rows, cols, physicalCols, startRow, containedRows);
	for (int i = 0;  i < containedRows * physicalCols; ++i)
		this->data[i] = 0;
}

PartialMatrix::PartialMatrix(int rows, int cols, int physicalCols, int startRow, int containedRows, float * data) {
	init(rows, cols, physicalCols, startRow, containedRows);
	for (int r = 0; r < containedRows; ++r) {
		for (int c = 0; c < cols; ++c)
			this->data[r * physicalCols + c] = data[r * cols + c];
		for (int c = cols; c < physicalCols; ++c)
			this->data[r * physicalCols + c] = 0;
	}
}

PartialMatrix::PartialMatrix(int rows, int cols, int physicalCols, int startRow, int containedRows, bool identity) {
	init(rows, cols, physicalCols, startRow, containedRows);
	for (int i = 0;  i < containedRows * physicalCols; ++i)
		this->data[i] = 0;
	if (identity && rows == cols) {
		for (int i = 0; i < containedRows; ++i) {
			data[i * physicalCols + (startRow + i)] = 1;
		}
	}
}

PartialMatrix::PartialMatrix(const PartialMatrix & source) {
	init(source.rows, source.cols, source.physicalCols, source.startRow, source.containedRows);
	for (int i = 0;  i < containedRows * physicalCols; ++i)
		this->data[i] = source.data[i];
}

PartialMatrix::~PartialMatrix() {
	free_align((void *)data);
}

void PartialMatrix::print() {
	printf("=== %d x %d: %d rows: [%d, %d) ===\n", rows, cols, containedRows, startRow, getRowLimit());
	for (int r = startRow; r < getRowLimit(); ++r) {
		printf("%f", (*this)(r, 0));
		for (int c = 1; c < cols; ++c)
			printf(" %f", (*this)(r, c));
		printf("\n");
	}
	printf("======\n");
}

void PartialMatrix::multiplyRow(int row, float factor) {
	if (factor == 1) return;
	vector float zero_v = (vector float){0.0f, 0.0f, 0.0f, 0.0f};
	vector float factor_v = (vector float){factor, factor, factor, factor};
	vector float * data_v = (vector float *)(data + (row - startRow) * physicalCols);
	for (int j = 0; j < physicalCols / VECTOR_WIDTH; ++j) {
		data_v[j] = spu_madd(data_v[j], factor_v, zero_v);
	}
}

void PartialMatrix::addRows(float factor, int sourceRow, int destinationRow) {
	if (factor == 0) return;
	vector float factor_v = (vector float){factor, factor, factor, factor};
	vector float * source_v = (vector float *)(data + (sourceRow - startRow) * physicalCols);
	vector float * destination_v = (vector float *)(data + (destinationRow - startRow) * physicalCols);
	for (int j = 0; j < physicalCols / VECTOR_WIDTH; ++j) {
		destination_v[j] = spu_madd(source_v[j], factor_v, destination_v[j]);
	}
}

// Adds a multiple of a row that is not a part of this PartialMatrix to one of the rows that are a part of it
void PartialMatrix::addRows(float factor, float * sourceRow, int destinationRow) {
	if (factor == 0) return;
	vector float factor_v = (vector float){factor, factor, factor, factor};
	vector float * source_v = (vector float *)sourceRow;
	vector float * destination_v = (vector float *)(data + (destinationRow - startRow) * physicalCols);
	for (int j = 0; j < physicalCols / VECTOR_WIDTH; ++j) {
		destination_v[j] = spu_madd(source_v[j], factor_v, destination_v[j]);
	}
}

void PartialMatrix::swapRows(int firstRow, int secondRow) {
	if (firstRow == secondRow) return;
	for (int j = 0; j < cols; ++j) {
		float tmp = (*this)(firstRow, j);
		(*this)(firstRow, j) = (*this)(secondRow, j);
		(*this)(secondRow, j) = tmp;
	}
}

