#include "SpuTableauSimplex.h"
#include "../types.h"
#include <spu_intrinsics.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <cstdio>

using namespace std;

void SpuTableauSimplex::pivot(PartialMatrix & tableau, float * pivotRow, int leaving, int entering) {
	float xFactor = pivotRow[entering];
	
	// Cancel out occurrences of the entering variable
	for (int i = tableau.getStartRow(); i < tableau.getRowLimit(); ++i) {
		if (i == leaving) continue;
		float factor = -tableau(i, entering) / xFactor;
		float savedColVal = tableau(i, entering);
		tableau.addRows(factor, pivotRow, i);
		tableau(i, entering) = savedColVal / xFactor;
	}
	
	if (leaving != -1) {
		tableau.multiplyRow(leaving, -1 / xFactor);
		tableau(leaving, entering) = 1 / xFactor;
	}
}

// Communications (* means "all";[^x] means "all except x"):
// Each communication end point is tagged in the code as "comm0" etc.
// 0. SPU0 -> PPU: entering variable, or optimality
// 1. PPU -> SPU[^0]: entering variable, or termination instruction
// 2. SPU* -> PPU: value and index of leaving variable, or unboundedness (both determined locally; PPU determines global choice - let x be the spu that 'wins')
// 3. PPU -> SPU*: whether this spu 'won', or termination instruction
// 4. SPUx -> PPU: transfer pivot row; spu writes to (ppe reads from) mbox to notify
// 5. PPU -> SPU[^x]: ppe writes to (spu reads from) mbox to sync; transfer pivot row
SimplexResult SpuTableauSimplex::iterate(PartialMatrix & tableau, float * ppuPivotRow, uint tagId, int spuIndex) {
	int n = tableau.getCols() - 1;
	int entering = -1;
	volatile float * pivotRow;
	static volatile float * incomingPivotRowBuffer = (float *)malloc_align(tableau.getPhysicalCols() * sizeof(float), ALIGN_QUAD_LOG2);
	
	if (tableau.getStartRow() == 0) {
		// Find entering variable by searching the objective function (row 0) for a positive coefficient (disregard the constant in column 0)
		for (int j = 1; j <= n; ++j) {
			// We are not using Bland's rule here, since the SPEs do not maintain the basic/nonbasic lists (they could of course be arranged to do so)
			if (tableau(0, j) > 0 && (entering == -1 || tableau(0, j) > tableau(0, entering)))
				entering = j;
		}
		if (entering == -1) {
			spu_write_out_mbox(SIMPLEX_MBOX_OPTIMAL); //comm0
			return OPTIMAL;
		}
		else
			spu_write_out_mbox((uint)entering); //comm0
	}
	else {
		entering = (int)spu_read_in_mbox(); //comm1
		if ((uint)entering == SIMPLEX_MBOX_OPTIMAL)
			return OPTIMAL;
	}
	
	// Find leaving variable by searching the column of the entering variable and determine the strictest bound
	int localLeaving = -1;
	Value32 largestRatio;
	int i = tableau.getStartRow();
	if (i == 0) // Skip objective function row - this also handles the case where this partial matrix only contains the objective function, in which case SIMPLEX_MBOX_UNBOUNDED will be returned by SPE 0
		i = 1;
	// We discovered a little too late that this part of the code is a little outdated - we are testing the inverse of the ratio that is described in the report, and therefore we must look for the largest ratio rather than the smallest. It is more cumbersome, but entirely equivalent to what is described in the report.
	for (; i < tableau.getRowLimit(); ++i) {
		float ratio;
		if (tableau(i, 0) == 0) {
			if (tableau(i, entering) == 0)
				ratio = 0;
			else if (tableau(i, entering) < 0)
				ratio = INFINITY;
			else
				ratio = -INFINITY;
		}
		else
			ratio = -tableau(i, entering) / tableau(i, 0);
		if (ratio <= 0) continue;
		if (localLeaving == -1 || ratio > largestRatio.floatValue) {
			largestRatio.floatValue = ratio;
			localLeaving = i;
		}
	}
	//comm2
	if (localLeaving == -1) {
		spu_write_out_mbox(SIMPLEX_MBOX_UNBOUNDED);
		spu_write_out_mbox((uint)-1);
	}
	else {
		spu_write_out_mbox(largestRatio.uintValue);
		spu_write_out_mbox((uint)localLeaving);
	}
		
	uint instruction = spu_read_in_mbox(); //comm3
	if (instruction == SIMPLEX_MBOX_UNBOUNDED)
		return UNBOUNDED;

	if (instruction == SIMPLEX_MBOX_LEAVING_IS_HERE) {
		//comm4 - pivot->ppu
		pivotRow = &tableau(localLeaving, 0);
		spu_mfcdma32((void *)pivotRow, (uint)ppuPivotRow, tableau.getPhysicalCols() * sizeof(float), tagId, MFC_PUT_CMD);
		(void)spu_mfcstat(MFC_TAG_UPDATE_ALL);
		spu_write_out_mbox(0U); //sync
	}
	else if (instruction == SIMPLEX_MBOX_LEAVING_IS_ELSEWHERE) {
		//comm5 - ppu->pivot
		pivotRow = incomingPivotRowBuffer;
		spu_read_in_mbox(); //sync
		spu_mfcdma32((void *)pivotRow, (uint)ppuPivotRow, tableau.getPhysicalCols() * sizeof(float), tagId, MFC_GET_CMD);
		(void)spu_mfcstat(MFC_TAG_UPDATE_ALL);
	}
	else {
		printf("Illegal instruction\n");
		return FAILURE;
	}
	
	pivot(tableau, (float *)pivotRow, instruction == SIMPLEX_MBOX_LEAVING_IS_HERE ? localLeaving : -1, entering);
	
	return SUBOPTIMAL;
}

