#ifndef MATRIX_H
#define MATRIX_H

#include <iostream>
#include <vector>
#include "gmpInterop.h"

class Matrix {
	friend std::ostream & operator << (std::ostream &, const Matrix &);
public:
	Matrix(int, int);
	Matrix(int rows, int cols, TYPE * data);
	Matrix(int, int, bool);
	Matrix(const Matrix &);
	~Matrix();
	int getRows() { return rows; }
	int getCols() { return cols; }
	int getPhysicalCols() { return physicalCols; }

#ifdef VOLATILE_DATA
	// When this class is being used in the parallel standard simplex method for Cell, the data buffer must be marked as volatile because it will be the target of DMA transfers
	volatile 
#endif
	TYPE * getData() { return data; }

	inline 
#ifdef VOLATILE_DATA
	volatile
#endif
	TYPE & operator () (int r, int c) { return data[r * physicalCols + c]; }

	inline TYPE operator () (int r, int c) const { return data[r * physicalCols + c]; }
	void swapRows(int firstRow, int secondRow);
	void multiplyRow(int row, TYPE factor);
	void addRows(int sourceRow, int destinationRow, TYPE factor);
	void print(const std::vector<int> & basic, const std::vector<int> & nonbasic);
private:
	int rows;
	int cols;
	int physicalCols;
#ifdef VOLATILE_DATA
	volatile
#endif
	TYPE * data;
};

std::ostream & operator << (std::ostream &, const Matrix &);

// Used by addRows to increment one number by another, with support for two different "zeroing rules", controlled by defines:
// - set to zero if result is below ZEROING_RULE_EPSILON
// - set to zero if the ratios of each of the operands to the result is greater than ZEROING_RULE_RATIO
// - not defining either of these will simply result in a regular a += b operation
#ifdef VOLATILE_DATA
inline void incr(volatile TYPE & a, const TYPE & b) {
#else
inline void incr(TYPE & a, const TYPE & b) {
#endif
	#ifdef ZEROING_RULE_EPSILON
		a += b;
		if (ABS(a) <= ZEROING_RULE_EPSILON) {
			a = 0;
		}
	#else
		#ifdef ZEROING_RULE_RATIO
			TYPE result = a + b;
			if (result == 0 || (ABS(a / result) >= ZEROING_RULE_RATIO && ABS(b / result) >= ZEROING_RULE_RATIO)) {
				a = 0;
			}
			else {
				a = result;
			}
		#else
			a += b;
		#endif
	#endif
}

#endif
