#include "Matrix.h"
#include "TableauSimplex.h"
#include "mps.h"
#include <cmath>
#include <iostream>
#include <vector>
#include <cstdlib>
#include <fstream>
#include <cstring>
#include <ctime>
#include "timer.h"

#define FEASIBILITY_THRESHOLD 1.0e-5 // abs(x0) must be below this value in order for the program to be declared feasible (only applies if Phase I is needed)

using namespace std;

int main(int argc, char * argv[]) {
	int rows, cols;
	bool initiallyFeasible = true;
	bool print = argc >= 3 && strcmp(argv[2], "print") == 0;
	istream * input;
	ifstream infile;
	
	if (argc == 1)
		input = &cin;
	else {
		infile.open(argv[1]);
		input = &infile;
	}
	vector<vector<TYPE> > parsedTableau = parse(*input);
	rows = parsedTableau.size();
	cols = parsedTableau[0].size();
	Matrix A(rows, cols + 1);
	for (int r = 0; r < rows; ++r) {
		for (int c = 0; c < cols; ++c) {
			A(r, c) = parsedTableau[r][c];
		}
		if (r > 0 && A(r, 0) < 0) {
			initiallyFeasible = false;
		}
		A(r, cols) = 1;
	}
	
	unsigned long long startTime = rdtsc();

	std::vector<int> basic, nonbasic;
	// Nonbasic variables are labeled 1 .. n
	for (int i = 1; i < cols; ++i)
		nonbasic.push_back(i);
	nonbasic.push_back(0); // Phase I variable
	// Basic variables are labeled n+1 .. n+m
	for (int i = cols; i < cols + rows - 1; ++i)
		basic.push_back(i);

	int numIterations = 0;
	SimplexResult status = SUBOPTIMAL;

	Matrix obj(1, cols); // Saves the original objective function
	if (!initiallyFeasible) {
		cout << "Entering Phase I" << endl;
		for (int c = 0; c < cols; ++c) {
			obj(0, c) = A(0, c);
			A(0, c) = 0;
		}
		A(0, cols) = -1; // The goal is to maximize -x0
		
		int leaving = 1;
		for (int i = 2; i < rows; ++i) {
			if (A(i, 0) < A(leaving, 0))
				leaving = i;
		}
		TableauSimplex::pivot(A, basic, nonbasic, leaving, cols);
		if (print) A.print(basic, nonbasic);
		while ((status = TableauSimplex::iteration(A, basic, nonbasic)) == SUBOPTIMAL) {
			++numIterations;
			if (print) A.print(basic, nonbasic);
			cout << numIterations << ": " << A(0, 0) << endl;
		}
		if (status == INFEASIBLE_OR_UNBOUNDED || ABS(A(0, 0)) > FEASIBILITY_THRESHOLD) {
			cout << "INFEASIBLE" << endl;
			return 0;
		}
		cout << "PHASE I COMPLETED" << endl;
		if (print) A.print(basic, nonbasic);
	}
	
	// Locate x0, which is expected to be nonbasic 
	int x0 = -1;
	for (int i = 0; i < cols; ++i) {
		if (nonbasic[i] == 0) {
			x0 = i + 1;
			nonbasic.erase(nonbasic.begin() + i);
			break;
		}
	}
	if (x0 == -1) {
		for (int j = 0; j < rows - 1; ++j) {
			if (basic[j] == 0) {
				x0 = j + 1;
				break;
			}
		}
		cout << "x0 is basic, and has value " << A(x0, 0) << " - terminating" << endl;
		// If x0 is basic, but has value 0, it should be possible to continue by pivoting it out, but we haven't spent time on this since the program usually does not give the right answer anyway
		return 0;
	}
	
	// Even if there was no Phase I, we still copy the tableau - this should be avoided
	Matrix newTableau(rows, cols);
	for (int i = 0; i < rows; ++i) {
		for (int j = 0; j < cols; ++j) {
			newTableau(i, j) = A(i, j < x0 ? j : j + 1);
		}
	}
	if (!initiallyFeasible) {
		if (print) newTableau.print(basic, nonbasic);
		newTableau(0, 0) = obj(0, 0);
		for (int j = 1; j < cols; ++j)
			if (nonbasic[j - 1] < cols)
				newTableau(0, j) = obj(0, nonbasic[j - 1]);
		for (int i = 1; i < rows; ++i) {
			if (basic[i - 1] < cols) {
				newTableau.addRows(i, 0, obj(0, basic[i - 1]));
			}
		}
	}
	
	if (print) newTableau.print(basic, nonbasic);
	cout << "Entering phase II" << endl;
	while ((status = TableauSimplex::iteration(newTableau, basic, nonbasic)) == SUBOPTIMAL) {
		++numIterations;
		if (print) newTableau.print(basic, nonbasic);
		cout << numIterations << ": " << newTableau(0, 0) << endl;
	}
	if (status == INFEASIBLE_OR_UNBOUNDED) {
		cout << "UNBOUNDED" << endl;
		return 0;
	}
	cout << "OPTIMAL" << endl;
	cout << "Optimal value: " << newTableau(0, 0) << endl;
	cout << "Elapsed time (minus input parsing): " << rdtsc() - startTime << endl;
	return 0;
}

